"""
    pyexcel_io.fileformat.tsvz
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level tsvz file format handler.

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.constants import FILE_FORMAT_TSVZ, KEYWORD_TSV_DIALECT

from .csvz import CSVZipBookWriter


class TSVZipBookWriter(CSVZipBookWriter):
    """ write zipped tsv file

    It is similiar to CSVZipBookWriter, but support tab separated values
    """

    def __init__(self):
        CSVZipBookWriter.__init__(self)
        self._file_type = FILE_FORMAT_TSVZ

    def open(self, file_name, **keywords):
        keywords["dialect"] = KEYWORD_TSV_DIALECT
        CSVZipBookWriter.open(self, file_name, **keywords)



"""
    pyexcel_io.constants
    ~~~~~~~~~~~~~~~~~~~

    Constants appeared in pyexcel

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License
"""
# flake8: noqa
DEFAULT_NAME = "pyexcel"
DEFAULT_SHEET_NAME = "%s_sheet1" % DEFAULT_NAME
DEFAULT_PLUGIN_NAME = "__%s_io_plugins__" % DEFAULT_NAME

MESSAGE_INVALID_PARAMETERS = "Invalid parameters"
MESSAGE_ERROR_02 = "No content, file name. Nothing is given"
MESSAGE_ERROR_03 = "cannot handle unknown content"
MESSAGE_WRONG_IO_INSTANCE = "Wrong io instance is passed for your file format."
MESSAGE_CANNOT_WRITE_STREAM_FORMATTER = (
    "Cannot write content of file type %s to stream"
)
MESSAGE_CANNOT_READ_STREAM_FORMATTER = (
    "Cannot read content of file type %s from stream"
)
MESSAGE_CANNOT_WRITE_FILE_TYPE_FORMATTER = (
    "Cannot write content of file type %s to file %s"
)
MESSAGE_CANNOT_READ_FILE_TYPE_FORMATTER = (
    "Cannot read content of file type %s from file %s"
)
MESSAGE_LOADING_FORMATTER = (
    "The plugin for file type %s is not installed. Please install %s"
)
MESSAGE_EMPTY_ARRAY = "One empty row is found"
MESSAGE_IGNORE_ROW = "One row is ignored"
MESSAGE_DB_EXCEPTION = """
Warning: Bulk insertion got below exception. Trying to do it one by one slowly."""

FILE_FORMAT_CSV = "csv"
FILE_FORMAT_TSV = "tsv"
FILE_FORMAT_CSVZ = "csvz"
FILE_FORMAT_TSVZ = "tsvz"
FILE_FORMAT_ODS = "ods"
FILE_FORMAT_XLS = "xls"
FILE_FORMAT_XLSX = "xlsx"
FILE_FORMAT_XLSM = "xlsm"
DB_SQL = "sql"
DB_DJANGO = "django"
KEYWORD_TSV_DIALECT = "excel-tab"
KEYWORD_LINE_TERMINATOR = "lineterminator"

SKIP_DATA = -1
TAKE_DATA = 0
STOP_ITERATION = 1


DEFAULT_MULTI_CSV_SEPARATOR = "__"
SEPARATOR_FORMATTER = "---%s---" % DEFAULT_NAME + "%s"
SEPARATOR_MATCHER = "---%s:(.*)---" % DEFAULT_NAME
DEFAULT_CSV_STREAM_FILE_FORMATTER = "---%s:" % DEFAULT_NAME + "%s---%s"
DEFAULT_CSV_NEWLINE = "\r\n"

MAX_INTEGER = 999999999999999



"""
    pyexcel_io.manager
    ~~~~~~~~~~~~~~~~~~~

    Control file streams

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io._compact import StringIO, BytesIO


MIME_TYPES = {}
FILE_TYPES = ()
TEXT_STREAM_TYPES = []
BINARY_STREAM_TYPES = []


def register_stream_type(file_type, stream_type):
    """
    keep track of stream type for different file formats
    """
    if stream_type == "text":
        TEXT_STREAM_TYPES.append(file_type)
    elif stream_type == "binary":
        BINARY_STREAM_TYPES.append(file_type)


def get_io(file_type):
    """A utility function to help you generate a correct io stream

    :param file_type: a supported file type
    :returns: a appropriate io stream, None otherwise
    """
    __file_type = None
    if file_type:
        __file_type = file_type.lower()

    if __file_type in TEXT_STREAM_TYPES:
        return StringIO()

    elif __file_type in BINARY_STREAM_TYPES:
        return BytesIO()

    else:
        return None


def get_io_type(file_type):
    """A utility function to help you generate a correct io stream

    :param file_type: a supported file type
    :returns: a appropriate io stream, None otherwise
    """
    __file_type = None
    if file_type:
        __file_type = file_type.lower()

    if __file_type in TEXT_STREAM_TYPES:
        return "string"

    elif __file_type in BINARY_STREAM_TYPES:
        return "bytes"

    else:
        return None


def register_a_file_type(file_type, stream_type, mime_type):
    """
    keep track of file format supports by this library
    """
    global FILE_TYPES
    FILE_TYPES += (file_type,)
    stream_type = stream_type
    if mime_type is not None:
        MIME_TYPES[file_type] = mime_type
    register_stream_type(file_type, stream_type)



"""
    pyexcel_io.database.querysets
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level handler for querysets

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import datetime
from itertools import chain

from pyexcel_io.sheet import SheetReader


class QuerysetsReader(SheetReader):
    """ turn querysets into an array """

    def __init__(self, query_sets, column_names, **keywords):
        SheetReader.__init__(self, query_sets, **keywords)
        self.__column_names = column_names
        self.__query_sets = query_sets

    def to_array(self):
        """
        Convert query sets into an array
        """
        if len(self.__query_sets) == 0:
            yield []

        for element in SheetReader.to_array(self):
            yield element

    def row_iterator(self):
        return chain([self.__column_names], self.__query_sets)

    def column_iterator(self, row):
        if self.__column_names is None:
            return

        if isinstance(row, list):
            for element in row:
                yield element

        else:
            for column in self.__column_names:
                if "__" in column:
                    value = get_complex_attribute(row, column)
                else:
                    value = get_simple_attribute(row, column)
                yield value


def get_complex_attribute(row, attribute):
    """ recursively get an attribute """
    attributes = attribute.split("__")
    value = row
    try:
        for attributee in attributes:
            value = get_simple_attribute(value, attributee)
    except AttributeError:
        value = None
    return value


def get_simple_attribute(row, attribute):
    """ get dotted attribute """
    value = getattr(row, attribute)
    if isinstance(value, (datetime.date, datetime.time)):
        value = value.isoformat()
    return value



"""
    pyexcel_io.writers.csvw
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level csv file format writer

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import csv
import codecs

from pyexcel_io.book import BookWriter
from pyexcel_io.sheet import SheetWriter
import pyexcel_io._compact as compact
import pyexcel_io.constants as constants


class UnicodeWriter(object):
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, file_handle, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = compact.StringIO()
        self.writer = csv.writer(self.queue, **kwds)
        self.stream = file_handle
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        """ write row into the csv file """
        self.writer.writerow(
            [compact.text_type(s).encode("utf-8") for s in row]
        )
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        """ write multiple rows into csv file """
        for row in rows:
            self.writerow(row)


class CSVSheetWriter(SheetWriter):
    """
    csv file writer

    """

    def __init__(
        self,
        filename,
        name,
        encoding="utf-8",
        single_sheet_in_book=False,
        sheet_index=None,
        **keywords
    ):
        self._encoding = encoding
        self._sheet_name = name
        self._single_sheet_in_book = single_sheet_in_book
        self.__line_terminator = constants.DEFAULT_CSV_NEWLINE
        if constants.KEYWORD_LINE_TERMINATOR in keywords:
            self.__line_terminator = keywords.get(
                constants.KEYWORD_LINE_TERMINATOR
            )
        if single_sheet_in_book:
            self._sheet_name = None
        self._sheet_index = sheet_index
        self.writer = None
        self.file_handle = None
        SheetWriter.__init__(
            self, filename, self._sheet_name, self._sheet_name, **keywords
        )

    def write_row(self, array):
        """
        write a row into the file
        """
        self.writer.writerow(array)


class CSVFileWriter(CSVSheetWriter):
    """ Write csv to a physical file """

    def close(self):
        self.file_handle.close()

    def set_sheet_name(self, name):
        if name != constants.DEFAULT_SHEET_NAME:
            names = self._native_book.split(".")
            file_name = "%s%s%s%s%s.%s" % (
                names[0],
                constants.DEFAULT_MULTI_CSV_SEPARATOR,
                name,  # sheet name
                constants.DEFAULT_MULTI_CSV_SEPARATOR,
                self._sheet_index,  # sheet index
                names[1],
            )
        else:
            file_name = self._native_book
        if compact.PY2:
            self.file_handle = open(file_name, "wb")
            self.writer = UnicodeWriter(
                self.file_handle, encoding=self._encoding, **self._keywords
            )
        else:
            self.file_handle = open(
                file_name, "w", newline="", encoding=self._encoding
            )
            self.writer = csv.writer(self.file_handle, **self._keywords)


class CSVMemoryWriter(CSVSheetWriter):
    """ Write csv to a memory stream """

    def __init__(
        self,
        filename,
        name,
        encoding="utf-8",
        single_sheet_in_book=False,
        sheet_index=None,
        **keywords
    ):
        CSVSheetWriter.__init__(
            self,
            filename,
            name,
            encoding=encoding,
            single_sheet_in_book=single_sheet_in_book,
            sheet_index=sheet_index,
            **keywords
        )

    def set_sheet_name(self, name):
        if compact.PY2:
            self.file_handle = self._native_book
            self.writer = UnicodeWriter(
                self.file_handle, encoding=self._encoding, **self._keywords
            )
        else:
            self.file_handle = self._native_book
            self.writer = csv.writer(self.file_handle, **self._keywords)
        if not self._single_sheet_in_book:
            self.writer.writerow(
                [
                    constants.DEFAULT_CSV_STREAM_FILE_FORMATTER
                    % (self._sheet_name, "")
                ]
            )

    def close(self):
        if self._single_sheet_in_book:
            #  on purpose, the this is not done
            #  because the io stream can be used later
            pass
        else:
            self.writer.writerow([constants.SEPARATOR_FORMATTER % ""])


class CSVBookWriter(BookWriter):
    """ write csv with unicode support """

    def __init__(self):
        BookWriter.__init__(self)
        self._file_type = constants.FILE_FORMAT_CSV
        self.__index = 0

    def create_sheet(self, name):
        writer_class = None
        if compact.is_string(type(self._file_alike_object)):
            writer_class = CSVFileWriter
        else:
            writer_class = CSVMemoryWriter
        writer = writer_class(
            self._file_alike_object,
            name,
            sheet_index=self.__index,
            **self._keywords
        )
        self.__index = self.__index + 1
        return writer



"""
    pyexcel_io.service
    ~~~~~~~~~~~~~~~~~~~

    provide service code to downstream projects

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import re
import math
import datetime

from pyexcel_io._compact import PY2
from pyexcel_io import constants
from pyexcel_io import exceptions


def has_no_digits_in_float(value):
    """check if a float value had zero value in digits"""
    return value == math.floor(value)


def detect_date_value(cell_text):
    """
    Read the date formats that were written by csv.writer
    """
    ret = None
    try:
        if len(cell_text) == 10:
            ret = datetime.datetime.strptime(cell_text, "%Y-%m-%d")
            ret = ret.date()
        elif len(cell_text) == 19:
            ret = datetime.datetime.strptime(cell_text, "%Y-%m-%d %H:%M:%S")
        elif len(cell_text) > 19:
            ret = datetime.datetime.strptime(
                cell_text[0:26], "%Y-%m-%d %H:%M:%S.%f"
            )
    except ValueError:
        pass
    return ret


def detect_float_value(
    cell_text, pep_0515_off=True, ignore_nan_text=False, default_float_nan=None
):
    should_we_skip_it = (
        cell_text.startswith("0") and cell_text.startswith("0.") is False
    )
    if should_we_skip_it:
        # do not convert if a number starts with 0
        # e.g. 014325
        return None

    if pep_0515_off:
        pattern = "([0-9]+_)+[0-9]+.[0-9]*$"
        if re.match(pattern, cell_text):
            return None

    try:
        if ignore_nan_text:
            if cell_text.lower() == "nan":
                return None
            else:
                return float(cell_text)
        else:
            if cell_text.lower() == "nan":
                if cell_text == default_float_nan:
                    return float("NaN")
                else:
                    return None
            else:
                return float(cell_text)
    except ValueError:
        return None


def detect_int_value(cell_text, pep_0515_off=True):
    if cell_text.startswith("0") and len(cell_text) > 1:
        return None

    if pep_0515_off:
        pattern = "([0-9]+_)+[0-9]+$"
        if re.match(pattern, cell_text):
            return None

    try:
        return int(cell_text)

    except ValueError:
        pattern = "([0-9]+,)*[0-9]+$"
        if re.match(pattern, cell_text):
            integer_string = cell_text.replace(",", "")
            return int(integer_string)

        else:
            return None


def float_value(value):
    """convert a value to float"""
    ret = float(value)
    return ret


def date_value(value):
    """convert to data value accroding ods specification"""
    ret = "invalid"
    try:
        # catch strptime exceptions only
        if len(value) == 10:
            ret = datetime.datetime.strptime(value, "%Y-%m-%d")
            ret = ret.date()
        elif len(value) == 19:
            ret = datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        elif len(value) > 19:
            ret = datetime.datetime.strptime(
                value[0:26], "%Y-%m-%dT%H:%M:%S.%f"
            )
    except ValueError:
        pass
    if ret == "invalid":
        raise Exception("Bad date value %s" % value)

    return ret


def time_value(value):
    """convert to time value accroding the specification"""
    import re

    results = re.match(r"PT(\d+)H(\d+)M(\d+)S", value)
    if results and len(results.groups()) == 3:
        hour = int(results.group(1))
        minute = int(results.group(2))
        second = int(results.group(3))
        if hour < 24:
            ret = datetime.time(hour, minute, second)
        else:
            ret = datetime.timedelta(
                hours=hour, minutes=minute, seconds=second
            )
    else:
        ret = None
    return ret


def boolean_value(value):
    """get bolean value"""
    if value == "true":
        ret = True
    elif value == "false":
        ret = False
    else:
        # needed for pyexcel-ods3
        ret = value
    return ret


ODS_FORMAT_CONVERSION = {
    "float": float,
    "date": datetime.date,
    "time": datetime.time,
    "timedelta": datetime.timedelta,
    "boolean": bool,
    "percentage": float,
    "currency": float,
}


ODS_WRITE_FORMAT_COVERSION = {
    float: "float",
    int: "float",
    str: "string",
    datetime.date: "date",
    datetime.time: "time",
    datetime.timedelta: "timedelta",
    bool: "boolean",
}

if PY2:
    ODS_WRITE_FORMAT_COVERSION[unicode] = "string"  # noqa: F821
    ODS_WRITE_FORMAT_COVERSION[long] = "throw_exception"  # noqa: F821


VALUE_CONVERTERS = {
    "float": float_value,
    "date": date_value,
    "time": time_value,
    "timedelta": time_value,
    "boolean": boolean_value,
    "percentage": float_value,
}


def throw_exception(value):
    raise exceptions.IntegerAccuracyLossError("%s is too big" % value)


def ods_float_value(value):
    if value > constants.MAX_INTEGER:
        raise exceptions.IntegerAccuracyLossError("%s is too big" % value)
    return value


def ods_date_value(value):
    return value.strftime("%Y-%m-%d")


def ods_time_value(value):
    return value.strftime("PT%HH%MM%SS")


def ods_bool_value(value):
    """convert a boolean value to text"""
    if value is True:
        return "true"

    else:
        return "false"


def ods_timedelta_value(cell):
    """convert a cell value to time delta"""
    hours = cell.days * 24 + cell.seconds // 3600
    minutes = (cell.seconds // 60) % 60
    seconds = cell.seconds % 60
    return "PT%02dH%02dM%02dS" % (hours, minutes, seconds)


ODS_VALUE_CONVERTERS = {
    "date": ods_date_value,
    "time": ods_time_value,
    "boolean": ods_bool_value,
    "timedelta": ods_timedelta_value,
    "float": ods_float_value,
    "throw_exception": throw_exception
}


VALUE_TOKEN = {
    "float": "value",
    "date": "date-value",
    "time": "time-value",
    "boolean": "boolean-value",
    "percentage": "value",
    "currency": "value",
    "timedelta": "time-value",
}



"""
    pyexcel_io.database.sql
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level handler for database import and export

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.book import BookWriter
from pyexcel_io.sheet import SheetWriter
from pyexcel_io.utils import is_empty_array, swap_empty_string_for_none
import pyexcel_io.constants as constants


class PyexcelSQLSkipRowException(Exception):
    """
    Raised this exception to skipping a row
    while data import
    """

    pass


class SQLTableWriter(SheetWriter):
    """Write to a table
    """

    def __init__(
        self, importer, adapter, auto_commit=True, bulk_size=1000, **keywords
    ):
        SheetWriter.__init__(
            self, importer, adapter, adapter.get_name(), **keywords
        )
        self.__auto_commit = auto_commit
        self.__count = 0
        self.__bulk_size = bulk_size

    def write_row(self, array):
        if is_empty_array(array):
            print(constants.MESSAGE_EMPTY_ARRAY)
        else:
            new_array = swap_empty_string_for_none(array)
            try:
                self._write_row(new_array)
            except PyexcelSQLSkipRowException:
                print(constants.MESSAGE_IGNORE_ROW)
                print(new_array)

    def _write_row(self, array):
        row = dict(zip(self._native_sheet.column_names, array))
        obj = None
        if self._native_sheet.row_initializer:
            # allow initinalizer to return None
            # if skipping is needed
            obj = self._native_sheet.row_initializer(row)
        if obj is None:
            obj = self._native_sheet.table()
            for name in self._native_sheet.column_names:
                if self._native_sheet.column_name_mapping_dict is not None:
                    key = self._native_sheet.column_name_mapping_dict[name]
                else:
                    key = name
                setattr(obj, key, row[name])
        self._native_book.session.add(obj)
        if self.__auto_commit and self.__bulk_size != float("inf"):
            self.__count += 1
            if self.__count % self.__bulk_size == 0:
                self._native_book.session.commit()

    def close(self):
        if self.__auto_commit:
            self._native_book.session.commit()


class SQLBookWriter(BookWriter):
    """ write data into database tables via sqlalchemy """

    def __init__(self):
        BookWriter.__init__(self)
        self.__importer = None
        self.__auto_commit = True

    def open_content(self, file_content, auto_commit=True, **keywords):
        self.__importer = file_content
        self.__auto_commit = auto_commit

    def create_sheet(self, sheet_name):
        sheet_writer = None
        adapter = self.__importer.get(sheet_name)
        if adapter:
            sheet_writer = SQLTableWriter(
                self.__importer, adapter, auto_commit=self.__auto_commit
            )
        else:
            raise Exception(
                "Sheet: %s does not match any given tables." % sheet_name
                + "Please be aware of case sensitivity."
            )

        return sheet_writer



"""
    pyexcel_io.io
    ~~~~~~~~~~~~~~~~~~~

    The io interface to file extensions

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from types import GeneratorType
import warnings

from pyexcel_io._compact import isstream, PY2
from pyexcel_io.plugins import READERS, WRITERS
import pyexcel_io.constants as constants


def iget_data(afile, file_type=None, **keywords):
    """Get data from an excel file source

    :param afile: a file name, a file stream or actual content
    :param sheet_name: the name of the sheet to be loaded
    :param sheet_index: the index of the sheet to be loaded
    :param sheets: a list of sheet to be loaded
    :param file_type: used only when filename is not a physial file name
    :param force_file_type: used only when filename refers to a physical file
                            and it is intended to open it as forced file type.
    :param streaming: toggles the type of returned data. The values of the
                      returned dictionary remain as generator if it is set
                      to True. Default is False.
    :param library: explicitly name a library for use.
                    e.g. library='pyexcel-ods'
    :param auto_detect_float: defaults to True
    :param auto_detect_int: defaults to True
    :param auto_detect_datetime: defaults to True
    :param ignore_infinity: defaults to True
    :param ignore_nan_text: various forms of 'NaN', 'nan' are ignored
    :param default_float_nan: choose one form of 'NaN', 'nan'
    :param pep_0515_off: turn off pep 0515. default to True.
    :param keywords: any other library specific parameters
    :returns: an ordered dictionary
    """
    data, reader = _get_data(
        afile, file_type=file_type, streaming=True, **keywords
    )
    return data, reader


def get_data(afile, file_type=None, streaming=None, **keywords):
    """Get data from an excel file source

    :param afile: a file name, a file stream or actual content
    :param sheet_name: the name of the sheet to be loaded
    :param sheet_index: the index of the sheet to be loaded
    :param file_type: used only when filename is not a physial file name
    :param streaming: toggles the type of returned data. The values of the
                      returned dictionary remain as generator if it is set
                      to True. Default is False.
    :param library: explicitly name a library for use.
                    e.g. library='pyexcel-ods'
    :param auto_detect_float: defaults to True
    :param auto_detect_int: defaults to True
    :param auto_detect_datetime: defaults to True
    :param ignore_infinity: defaults to True
    :param keywords: any other library specific parameters
    :returns: an ordered dictionary
    """
    if streaming is not None and streaming is True:
        warnings.warn("Please use iget_data instead")
    data, _ = _get_data(
        afile, file_type=file_type, streaming=False, **keywords
    )
    return data


def _get_data(afile, file_type=None, **keywords):
    if isstream(afile):
        keywords.update(
            dict(
                file_stream=afile,
                file_type=file_type or constants.FILE_FORMAT_CSV,
            )
        )
    else:
        if afile is None or file_type is None:
            keywords.update(dict(file_name=afile, file_type=file_type))
        else:
            keywords.update(dict(file_content=afile, file_type=file_type))
    return load_data(**keywords)


def save_data(afile, data, file_type=None, **keywords):
    """Save data to an excel file source

    Your data must be a dictionary

    :param filename: actual file name, a file stream or actual content
    :param data: a dictionary but an ordered dictionary is preferred
    :param file_type: used only when filename is not a physial file name
    :param library: explicitly name a library for use.
                    e.g. library='pyexcel-ods'
    :param keywords: any other parameters that python csv module's
                     `fmtparams <https://docs.python.org/release/3.1.5/library/csv.html#dialects-and-formatting-parameters>`_
    """  # noqa
    to_store = data

    is_list = isinstance(data, (list, GeneratorType))
    if is_list:
        single_sheet_in_book = True
        to_store = {constants.DEFAULT_SHEET_NAME: data}
    else:
        if PY2:
            keys = data.keys()
        else:
            keys = list(data.keys())
        single_sheet_in_book = len(keys) == 1

    no_file_type = isstream(afile) and file_type is None
    if no_file_type:
        file_type = constants.FILE_FORMAT_CSV

    store_data(
        afile,
        to_store,
        file_type=file_type,
        single_sheet_in_book=single_sheet_in_book,
        **keywords
    )


def store_data(afile, data, file_type=None, **keywords):
    """Non public function to store data to afile

    :param filename: actual file name, a file stream or actual content
    :param data: the data to be written
    :param file_type: used only when filename is not a physial file name
    :param keywords: any other parameters
    """
    if isstream(afile):
        keywords.update(dict(file_stream=afile, file_type=file_type))
    else:
        keywords.update(dict(file_name=afile, file_type=file_type))
    with get_writer(**keywords) as writer:
        writer.write(data)


def load_data(
    file_name=None,
    file_content=None,
    file_stream=None,
    file_type=None,
    force_file_type=None,
    sheet_name=None,
    sheet_index=None,
    sheets=None,
    library=None,
    streaming=False,
    **keywords
):
    """Load data from any supported excel formats

    :param filename: actual file name, a file stream or actual content
    :param file_type: used only when filename is not a physial file name
    :param force_file_type: used only when filename refers to a physical file
                            and it is intended to open it as forced file type.
    :param sheet_name: the name of the sheet to be loaded
    :param sheet_index: the index of the sheet to be loaded
    :param keywords: any other parameters
    """
    result = {}
    inputs = [file_name, file_content, file_stream]
    number_of_none_inputs = [x for x in inputs if x is not None]
    if len(number_of_none_inputs) != 1:
        raise IOError(constants.MESSAGE_ERROR_02)

    if file_type is None:
        if force_file_type:
            file_type = force_file_type
        else:
            try:
                file_type = file_name.split(".")[-1]
            except AttributeError:
                raise Exception("file_name should be a string type")

    reader = READERS.get_a_plugin(file_type, library)
    if file_name:
        reader.open(file_name, **keywords)
    elif file_content:
        reader.open_content(file_content, **keywords)
    elif file_stream:
        reader.open_stream(file_stream, **keywords)
    if sheet_name:
        result = reader.read_sheet_by_name(sheet_name)
    elif sheet_index is not None:
        result = reader.read_sheet_by_index(sheet_index)
    elif sheets is not None:
        result = reader.read_many(sheets)
    else:
        result = reader.read_all()
    if streaming is False:
        for key in result.keys():
            result[key] = list(result[key])
        reader.close()
        reader = None

    return result, reader


def get_writer(
    file_name=None, file_stream=None, file_type=None, library=None, **keywords
):
    """find a suitable writer"""
    inputs = [file_name, file_stream]
    number_of_none_inputs = [x for x in inputs if x is not None]

    if len(number_of_none_inputs) != 1:
        raise IOError(constants.MESSAGE_ERROR_02)

    file_type_given = True
    if file_type is None and file_name:
        try:
            file_type = file_name.split(".")[-1]
        except AttributeError:
            raise Exception("file_name should be a string type")

        file_type_given = False

    writer = WRITERS.get_a_plugin(file_type, library)
    if file_name:
        if file_type_given:
            writer.open_content(file_name, **keywords)
        else:
            writer.open(file_name, **keywords)
    elif file_stream:
        writer.open_stream(file_stream, **keywords)
    # else: is resolved by earlier raise statement
    return writer



"""
    pyexcel_io.sheet
    ~~~~~~~~~~~~~~~~~~~

    The io interface to file extensions

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io._compact import irange
from pyexcel_io.utils import _index_filter
import pyexcel_io.constants as constants


class NamedContent(object):
    """
    Helper class for content that does not have a name
    """

    def __init__(self, name, payload):
        self.name = name
        self.payload = payload


class SheetReader(object):
    """
    Generic sheet reader
    """

    def __init__(
        self,
        sheet,
        start_row=0,
        row_limit=-1,
        start_column=0,
        column_limit=-1,
        skip_row_func=None,
        skip_column_func=None,
        skip_empty_rows=False,
        row_renderer=None,
        **keywords
    ):
        self._native_sheet = sheet
        self._keywords = {}
        self._keywords.update(keywords)
        self._start_row = start_row
        self._row_limit = row_limit
        self._start_column = start_column
        self._column_limit = column_limit
        self._skip_row = _index_filter
        self._skip_column = _index_filter
        self._skip_empty_rows = skip_empty_rows
        self._row_renderer = row_renderer

        if skip_row_func:
            self._skip_row = skip_row_func
        if skip_column_func:
            self._skip_column = skip_column_func

    def to_array(self):
        """2 dimentional representation of the content
        """
        for row_index, row in enumerate(self.row_iterator()):
            row_position = self._skip_row(
                row_index, self._start_row, self._row_limit
            )
            if row_position == constants.SKIP_DATA:
                continue

            elif row_position == constants.STOP_ITERATION:
                break

            return_row = []
            tmp_row = []

            for column_index, cell_value in enumerate(
                self.column_iterator(row)
            ):
                column_position = self._skip_column(
                    column_index, self._start_column, self._column_limit
                )
                if column_position == constants.SKIP_DATA:
                    continue

                elif column_position == constants.STOP_ITERATION:
                    break

                tmp_row.append(cell_value)
                if cell_value is not None and cell_value != "":
                    return_row += tmp_row
                    tmp_row = []
            if self._skip_empty_rows and len(return_row) < 1:
                # we by-pass next yeild here
                # because it is an empty row
                continue

            if self._row_renderer:
                return_row = self._row_renderer(return_row)
            yield return_row

    def row_iterator(self):
        """
        iterate each row

        override this function in the sitation where
        number_of_rows() is difficult or costly to implement
        """
        return irange(self.number_of_rows())

    def column_iterator(self, row):
        """
        iterate each column of a given row

        override this function in the sitation where
        number_of_columns() is difficult or costly to implement
        """
        for column in irange(self.number_of_columns()):
            yield self.cell_value(row, column)

    def number_of_rows(self):
        """
        implement this method for easy extension
        """
        raise Exception("Please implement number_of_rows()")

    def number_of_columns(self):
        """
        implement this method for easy extension
        """
        raise Exception("Please implement number_of_columns()")

    def cell_value(self, row, column):
        """
        implement this method for easy extension
        """
        raise Exception("Please implement cell_value()")

    def close(self):
        pass


class SheetWriter(object):
    """
    Generic sheet writer
    """

    def __init__(self, native_book, native_sheet, name, **keywords):
        if name:
            sheet_name = name
        else:
            sheet_name = constants.DEFAULT_SHEET_NAME
        self._native_book = native_book
        self._native_sheet = native_sheet
        self._keywords = keywords
        self.set_sheet_name(sheet_name)

    def set_sheet_name(self, name):
        """
        Set sheet name
        """
        pass

    def write_row(self, array):
        """
        write a row into the file
        """
        raise NotImplementedError("Please implement write_row()")

    def write_array(self, table):
        """
        For standalone usage, write an array
        """
        for row in table:
            self.write_row(row)

    def close(self):
        """
        This call actually save the file
        """
        pass



"""
    pyexcel_io.database.django
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level handler for django import and export

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.database.common import DbExporter
from pyexcel_io.database.querysets import QuerysetsReader


class DjangoModelReader(QuerysetsReader):
    """Read from django model
    """

    def __init__(self, model, export_columns=None, **keywords):
        self.__model = model
        if export_columns:
            column_names = export_columns
        else:
            column_names = sorted(
                [field.attname for field in self.__model._meta.concrete_fields]
            )
        QuerysetsReader.__init__(
            self, self.__model.objects.all(), column_names, **keywords
        )


class DjangoBookReader(DbExporter):
    """ read django models """

    def __init__(self):
        DbExporter.__init__(self)
        self.exporter = None

    def export_tables(self, file_content, **keywords):
        self.exporter = file_content
        self._load_from_django_models()

    def read_sheet(self, native_sheet):
        reader = DjangoModelReader(
            native_sheet.model, native_sheet.export_columns
        )
        return reader.to_array()

    def _load_from_django_models(self):
        self._native_book = self.exporter.adapters



import os
import datetime
from unittest import TestCase
from textwrap import dedent

import pyexcel as pe
from pyexcel_io._compact import text_type


class TestDateFormat(TestCase):
    def setUp(self):
        self.excel_filename = "testdateformat.csv"
        self.data = [
            [
                "01/06/2016",
                datetime.date(2014, 12, 25),
                datetime.datetime(2014, 12, 25, 11, 11, 11),
                datetime.datetime(2014, 12, 25, 11, 11, 11, 10),
            ]
        ]
        pe.save_as(dest_file_name=self.excel_filename, array=self.data)

    def test_auto_detect_float(self):
        sheet = pe.get_sheet(file_name=self.excel_filename)
        self.assertEqual(sheet.to_array(), self.data)

    def test_auto_detect_float_false(self):
        expected = [
            [
                "01/06/2016",
                "2014-12-25",
                "2014-12-25 11:11:11",
                "2014-12-25 11:11:11.000010",
            ]
        ]
        sheet = pe.get_sheet(
            file_name=self.excel_filename, auto_detect_datetime=False
        )
        self.assertEqual(sheet.to_array(), expected)

    def tearDown(self):
        os.unlink(self.excel_filename)


class TestAutoDetectInt(TestCase):
    def setUp(self):
        self.content = [[1, 2, 3.1]]
        self.test_file = "test_auto_detect_init.csv"
        pe.save_as(array=self.content, dest_file_name=self.test_file)

    def test_auto_detect_int(self):
        sheet = pe.get_sheet(file_name=self.test_file)
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +---+---+-----+
        | 1 | 2 | 3.1 |
        +---+---+-----+"""
        ).strip()
        self.assertEqual(str(sheet), expected)

    def test_get_book_auto_detect_int(self):
        book = pe.get_book(file_name=self.test_file)
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +---+---+-----+
        | 1 | 2 | 3.1 |
        +---+---+-----+"""
        ).strip()
        self.assertEqual(str(book), expected)

    def test_auto_detect_int_false(self):
        sheet = pe.get_sheet(file_name=self.test_file, auto_detect_int=False)
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +-----+-----+-----+
        | 1.0 | 2.0 | 3.1 |
        +-----+-----+-----+"""
        ).strip()
        self.assertEqual(str(sheet), expected)

    def test_get_book_auto_detect_int_false(self):
        book = pe.get_book(file_name=self.test_file, auto_detect_int=False)
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +-----+-----+-----+
        | 1.0 | 2.0 | 3.1 |
        +-----+-----+-----+"""
        ).strip()
        self.assertEqual(str(book), expected)

    def tearDown(self):
        os.unlink(self.test_file)


class TestAutoDetectFloat(TestCase):
    """
    csv format would keep all as text 1, 2.0, 3.1
    without auto-detection, they all stay as text format after reading
    out
    """

    def setUp(self):
        self.content = [[1, 2.0, 3.1]]
        self.test_file = "test_auto_detect_init.csv"
        pe.save_as(array=self.content, dest_file_name=self.test_file)

    def test_auto_detect_float_false(self):
        sheet = pe.get_sheet(file_name=self.test_file, auto_detect_float=False)
        self.assertEqual(sheet.to_array(), [[1, "2.0", "3.1"]])
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +---+-----+-----+
        | 1 | 2.0 | 3.1 |
        +---+-----+-----+"""
        ).strip()
        self.assertEqual(str(sheet), expected)

    def test_get_book_auto_detect_float_false(self):
        book = pe.get_book(file_name=self.test_file, auto_detect_float=False)
        self.assertEqual(book[0].to_array(), [[1, "2.0", "3.1"]])
        expected = dedent(
            """
        test_auto_detect_init.csv:
        +---+-----+-----+
        | 1 | 2.0 | 3.1 |
        +---+-----+-----+"""
        ).strip()
        self.assertEqual(str(book), expected)

    def tearDown(self):
        os.unlink(self.test_file)


class TestSpecialStrings(TestCase):
    """
    csv format would keep all as text 1, 2.0, 3.1
    without auto-detection, they all stay as text format after reading
    out
    """

    def setUp(self):
        self.content = [["01", 1, 2.0, 3.1, "NaN", "nan"]]
        self.test_file = "test_auto_detect_init.csv"
        pe.save_as(array=self.content, dest_file_name=self.test_file)

    def test_auto_detect_float_true(self):
        sheet = pe.get_sheet(file_name=self.test_file)
        self.assertEqual(sheet.to_array(), [["01", 1, 2, 3.1, "NaN", "nan"]])

    def test_auto_detect_float_false(self):
        sheet = pe.get_sheet(file_name=self.test_file, auto_detect_float=False)
        self.assertEqual(
            sheet.to_array(), [["01", 1, "2.0", "3.1", "NaN", "nan"]]
        )

    def test_auto_detect_float_ignore_nan_text(self):
        sheet = pe.get_sheet(file_name=self.test_file, ignore_nan_text=True)
        self.assertEqual(sheet.to_array(), [["01", 1, 2.0, 3.1, "NaN", "nan"]])

    def test_auto_detect_float_default_float_nan(self):
        sheet = pe.get_sheet(file_name=self.test_file, default_float_nan="nan")
        result = sheet.to_array()
        assert isinstance(result[0][5], float)
        assert isinstance(result[0][4], text_type)

    def tearDown(self):
        os.unlink(self.test_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from unittest import TestCase
from textwrap import dedent
from nose.tools import raises, eq_
import pyexcel_io.manager as manager
from pyexcel_io.sheet import NamedContent
from pyexcel_io.readers.csvr import (
    CSVSheetReader,
    CSVFileReader,
    CSVinMemoryReader,
)
from pyexcel_io.writers.csvw import CSVFileWriter, CSVMemoryWriter
from pyexcel_io._compact import BytesIO, PY2, StringIO


class TestReaders(TestCase):
    def setUp(self):
        self.file_type = "csv"
        self.test_file = "csv_book." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.expected_data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        with open(self.test_file, "w") as f:
            for row in self.data:
                f.write(",".join(row) + "\n")

    @raises(NotImplementedError)
    def test_sheet_reader(self):
        sheet = CSVSheetReader(self.test_file)
        sheet.get_file_handle()

    def test_sheet_file_reader(self):
        r = CSVFileReader(NamedContent(self.file_type, self.test_file))
        result = list(r.to_array())
        self.assertEqual(result, self.expected_data)

    def test_sheet_memory_reader(self):
        io = manager.get_io(self.file_type)
        with open(self.test_file, "r") as f:
            io.write(f.read())
        io.seek(0)
        r = CSVinMemoryReader(NamedContent(self.file_type, io))
        result = list(r.to_array())
        self.assertEqual(result, self.expected_data)

    def tearDown(self):
        os.unlink(self.test_file)


class TestWriter(TestCase):
    def setUp(self):
        self.file_type = "csv"
        self.test_file = "csv_book." + self.file_type
        self.data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.result = dedent(
            """
           1,2,3
           4,5,6
           7,8,9
        """
        ).strip("\n")

    def test_sheet_writer(self):
        w = CSVFileWriter(self.test_file, None)
        for row in self.data:
            w.write_row(row)
        w.close()
        with open(self.test_file, "r") as f:
            content = f.read().replace("\r", "")
            self.assertEqual(content.strip("\n"), self.result)

    def tearDown(self):
        os.unlink(self.test_file)


class TestMemoryWriter(TestCase):
    def setUp(self):
        self.file_type = "csv"
        self.test_file = "csv_book." + self.file_type
        self.data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.result = dedent(
            """
           1,2,3
           4,5,6
           7,8,9
        """
        ).strip("\n")

    def test_sheet_writer_to_memory(self):
        io = manager.get_io(self.file_type)
        w = CSVMemoryWriter(io, None, single_sheet_in_book=True)
        for row in self.data:
            w.write_row(row)
        w.close()
        content = io.getvalue().replace("\r", "")
        self.assertEqual(content.strip("\n"), self.result)


class TestNonUniformCSV(TestCase):
    def setUp(self):
        self.file_type = "csv"
        self.test_file = "csv_book." + self.file_type
        self.data = [["1"], ["4", "5", "6", "", ""], ["", "7"]]
        with open(self.test_file, "w") as f:
            for row in self.data:
                f.write(",".join(row) + "\n")

    def test_sheet_file_reader(self):
        r = CSVFileReader(NamedContent(self.file_type, self.test_file))
        result = list(r.to_array())
        self.assertEqual(result, [[1], [4, 5, 6], ["", 7]])

    def tearDown(self):
        os.unlink(self.test_file)


def test_utf16_decoding():
    test_file = os.path.join("tests", "fixtures", "csv-encoding-utf16.csv")
    reader = CSVFileReader(NamedContent("csv", test_file), encoding="utf-16")
    content = list(reader.to_array())
    if PY2:
        content[0] = [s.encode("utf-8") for s in content[0]]
    expected = [["Äkkilähdöt", "Matkakirjoituksia", "Matkatoimistot"]]
    eq_(content, expected)


def test_utf16_encoding():
    content = [[u"Äkkilähdöt", u"Matkakirjoituksia", u"Matkatoimistot"]]
    test_file = "test-utf16-encoding.csv"
    writer = CSVFileWriter(
        test_file, None, encoding="utf-16", lineterminator="\n"
    )
    writer.write_array(content)
    writer.close()
    with open(test_file, "rb") as f:
        actual = f.read().decode("utf-16")
        if PY2:
            actual = actual.encode("utf-8")
        eq_(actual, "Äkkilähdöt,Matkakirjoituksia,Matkatoimistot\n")
    os.unlink(test_file)


def test_utf16_memory_decoding():
    test_content = u"Äkkilähdöt,Matkakirjoituksia,Matkatoimistot"
    test_content = BytesIO(test_content.encode("utf-16"))
    reader = CSVinMemoryReader(
        NamedContent("csv", test_content), encoding="utf-16"
    )
    content = list(reader.to_array())
    if PY2:
        content[0] = [s.encode("utf-8") for s in content[0]]
    expected = [["Äkkilähdöt", "Matkakirjoituksia", "Matkatoimistot"]]
    eq_(content, expected)


def test_utf16_memory_encoding():
    content = [[u"Äkkilähdöt", u"Matkakirjoituksia", u"Matkatoimistot"]]
    io = StringIO()
    writer = CSVMemoryWriter(
        io,
        None,
        lineterminator="\n",
        single_sheet_in_book=True,
        encoding="utf-16",
    )
    writer.write_array(content)
    actual = io.getvalue()
    if PY2:
        actual = actual.decode("utf-16")
    eq_(actual, u"Äkkilähdöt,Matkakirjoituksia,Matkatoimistot\n")



from nose.tools import raises
from pyexcel_io.book import RWInterface, BookReader, BookWriter


@raises(NotImplementedError)
def test_rwinterface():
    interface = RWInterface()
    interface.open(1)


@raises(NotImplementedError)
def test_rwinterface_2():
    interface = RWInterface()
    interface.open_stream(1)


@raises(NotImplementedError)
def test_book_reader():
    reader = BookReader()
    reader.read_sheet(1)


@raises(IOError)
def test_book_reader_open_stream():
    reader = BookReader()
    reader.open_stream("a string")


@raises(IOError)
def test_book_writer():
    writer = BookWriter()
    writer.open_stream("a string")



"""
    pyexcel_io.database
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    database data importer and exporter

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.plugins import IOPluginInfoChain
from pyexcel_io.constants import DB_DJANGO, DB_SQL


IOPluginInfoChain(__name__).add_a_reader(
    relative_plugin_class_path="exporters.django.DjangoBookReader",
    file_types=[DB_DJANGO],
).add_a_reader(
    relative_plugin_class_path="exporters.sqlalchemy.SQLBookReader",
    file_types=[DB_SQL],
).add_a_writer(
    relative_plugin_class_path="importers.django.DjangoBookWriter",
    file_types=[DB_DJANGO],
).add_a_writer(
    relative_plugin_class_path="importers.sqlalchemy.SQLBookWriter",
    file_types=[DB_SQL],
)



"""
    pyexcel_io._compact
    ~~~~~~~~~~~~~~~~~~~

    Compatibles

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
# flake8: noqa
# pylint: disable=import-error
# pylint: disable=invalid-name
# pylint: disable=too-few-public-methods
# pylint: disable=ungrouped-imports
# pylint: disable=redefined-variable-type
import sys
import types
import logging

PY2 = sys.version_info[0] == 2
PY3_ABOVE = sys.version_info[0] >= 3
PY26 = PY2 and sys.version_info[1] < 7
PY27 = PY2 and sys.version_info[1] == 7
PY27_ABOVE = PY27 or PY3_ABOVE

if PY26:
    from ordereddict import OrderedDict
else:
    from collections import OrderedDict

try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


if PY2:
    from cStringIO import StringIO
    from cStringIO import StringIO as BytesIO

    text_type = unicode
    irange = xrange

    class Iterator(object):
        def next(self):
            return type(self).__next__(self)


else:
    from io import StringIO, BytesIO

    text_type = str
    Iterator = object
    irange = range


def isstream(instance):
    """ check if a instance is a stream """
    try:
        import mmap

        i_am_not_mmap_obj = not isinstance(instance, mmap.mmap)
    except ImportError:
        # Python 2.6 or Google App Engine
        i_am_not_mmap_obj = True

    return hasattr(instance, "read") and i_am_not_mmap_obj


def is_string(atype):
    """find out if a type is str or not"""
    if atype == str:
        return True

    elif PY2:
        if atype == unicode:
            return True

    return False



import os

from pyexcel_io import get_data, save_data
from pyexcel_io.utils import _index_filter
from nose.tools import eq_
import pyexcel_io.constants as constants


def test_index_filter():
    current_index, start, limit, expected = (0, 1, -1, constants.SKIP_DATA)
    eq_(_index_filter(current_index, start, limit), expected)
    current_index, start, limit, expected = (2, 1, -1, constants.TAKE_DATA)
    eq_(_index_filter(current_index, start, limit), expected)
    current_index, start, limit, expected = (2, 1, 10, constants.TAKE_DATA)
    eq_(_index_filter(current_index, start, limit), expected)
    current_index, start, limit, expected = (
        100,
        1,
        10,
        constants.STOP_ITERATION,
    )
    eq_(_index_filter(current_index, start, limit), expected)


class TestFilter:
    def setUp(self):
        self.test_file = "test_filter.csv"
        sample = [
            [1, 21, 31],
            [2, 22, 32],
            [3, 23, 33],
            [4, 24, 34],
            [5, 25, 35],
            [6, 26, 36],
        ]
        save_data(self.test_file, sample)

    def test_filter_row(self):
        filtered_data = get_data(self.test_file, start_row=3)
        expected = [[4, 24, 34], [5, 25, 35], [6, 26, 36]]
        eq_(filtered_data[self.test_file], expected)

    def test_filter_row_2(self):
        filtered_data = get_data(self.test_file, start_row=3, row_limit=1)
        expected = [[4, 24, 34]]
        eq_(filtered_data[self.test_file], expected)

    def test_filter_column(self):
        filtered_data = get_data(self.test_file, start_column=1)
        expected = [[21, 31], [22, 32], [23, 33], [24, 34], [25, 35], [26, 36]]
        eq_(filtered_data[self.test_file], expected)

    def test_filter_column_2(self):
        filtered_data = get_data(
            self.test_file, start_column=1, column_limit=1
        )
        expected = [[21], [22], [23], [24], [25], [26]]
        eq_(filtered_data[self.test_file], expected)

    def test_filter_both_ways(self):
        filtered_data = get_data(self.test_file, start_column=1, start_row=3)
        expected = [[24, 34], [25, 35], [26, 36]]
        eq_(filtered_data[self.test_file], expected)

    def test_filter_both_ways_2(self):
        filtered_data = get_data(
            self.test_file,
            start_column=1,
            column_limit=1,
            start_row=3,
            row_limit=1,
        )
        expected = [[24]]
        eq_(filtered_data[self.test_file], expected)

    def tearDown(self):
        os.unlink(self.test_file)



"""
    pyexcel_io.book
    ~~~~~~~~~~~~~~~~~~~

    The io interface to file extensions

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import pyexcel_io.manager as manager
from pyexcel_io._compact import OrderedDict, isstream, PY2
from .constants import MESSAGE_ERROR_03, MESSAGE_WRONG_IO_INSTANCE


class RWInterface(object):
    """
    The common methods for book reader and writer
    """

    stream_type = None

    def __init__(self):
        self._file_type = None

    def open(self, file_name, **keywords):
        """open a file for read or write"""
        raise NotImplementedError("Please implement this method")

    def open_stream(self, file_stream, **keywords):
        """open a file stream for read or write"""
        raise NotImplementedError("Please implement this method")

    def open_content(self, file_stream, **keywords):
        """open a file content for read or write"""
        raise NotImplementedError("Please implement this method")

    def set_type(self, file_type):
        """
        set the file type for the instance

        file type is needed when a third party library could
        handle more than one file type"""
        self._file_type = file_type

    def close(self):
        """
        close the file handle if necessary
        """
        pass

    # implement context manager

    def __enter__(self):
        return self

    def __exit__(self, a_type, value, traceback):
        self.close()


class BookReader(RWInterface):
    """
    Standard book reader
    """

    def __init__(self):
        super(BookReader, self).__init__()
        self._file_name = None
        self._file_stream = None
        self._keywords = None
        self._native_book = None

    def open(self, file_name, **keywords):
        """
        open a file with unlimited keywords

        keywords are passed on to individual readers
        """
        self._file_name = file_name
        self._keywords = keywords

    def open_stream(self, file_stream, **keywords):
        """
        open a file with unlimited keywords for reading

        keywords are passed on to individual readers
        """
        if isstream(file_stream):
            if PY2:
                if hasattr(file_stream, "seek"):
                    file_stream.seek(0)
                else:
                    # python 2
                    # Hei zipfile in odfpy would do a seek
                    # but stream from urlib cannot do seek
                    file_stream = _convert_content_to_stream(
                        file_stream.read(), self._file_type
                    )
            else:
                from io import UnsupportedOperation

                try:
                    file_stream.seek(0)
                except UnsupportedOperation:
                    # python 3
                    file_stream = _convert_content_to_stream(
                        file_stream.read(), self._file_type
                    )

            self._file_stream = file_stream
            self._keywords = keywords
        else:
            raise IOError(MESSAGE_WRONG_IO_INSTANCE)

    def open_content(self, file_content, **keywords):
        """
        read file content as if it is a file stream with
        unlimited keywords for reading

        keywords are passed on to individual readers
        """
        file_stream = _convert_content_to_stream(file_content, self._file_type)
        self.open_stream(file_stream, **keywords)

    def read_sheet_by_name(self, sheet_name):
        """
        read a named sheet from a excel data book
        """
        named_contents = [
            content
            for content in self._native_book
            if content.name == sheet_name
        ]
        if len(named_contents) == 1:
            return {named_contents[0].name: self.read_sheet(named_contents[0])}

        else:
            raise ValueError("Cannot find sheet %s" % sheet_name)

    def read_sheet_by_index(self, sheet_index):
        """
        read an indexed sheet from a excel data book
        """
        try:
            sheet = self._native_book[sheet_index]
            return {sheet.name: self.read_sheet(sheet)}

        except IndexError:
            self.close()
            raise

    def read_all(self):
        """
        read everything from a excel data book
        """
        result = OrderedDict()
        for sheet in self._native_book:
            result[sheet.name] = self.read_sheet(sheet)
        return result

    def read_many(self, sheets):
        """
        read everything from a excel data book
        """
        result = OrderedDict()
        for sheet in sheets:
            if isinstance(sheet, int):
                result.update(self.read_sheet_by_index(sheet))
            else:
                result.update(self.read_sheet_by_name(sheet))
        return result

    def read_sheet(self, native_sheet):
        """
        Return a context specific sheet from a native sheet
        """
        raise NotImplementedError("Please implement this method")


class BookWriter(RWInterface):
    """
    Standard book writer
    """

    def __init__(self):
        super(BookWriter, self).__init__()
        self._file_alike_object = None
        self._keywords = None

    def open(self, file_name, **keywords):
        """
        open a file with unlimited keywords for writing

        keywords are passed on to individual writers
        """
        self._file_alike_object = file_name
        self._keywords = keywords

    def open_stream(self, file_stream, **keywords):
        """
        open a file stream with unlimited keywords for writing

        keywords are passed on to individual writers
        """
        if not isstream(file_stream):
            raise IOError(MESSAGE_ERROR_03)

        self.open(file_stream, **keywords)

    def open_content(self, file_stream, **keywords):
        """open a file content for read or write"""
        raise Exception("Normal writer would not need this interface")

    def write(self, incoming_dict):
        """
        write a dictionary into an excel file
        """
        for sheet_name in incoming_dict:
            sheet_writer = self.create_sheet(sheet_name)
            if sheet_writer:
                sheet_writer.write_array(incoming_dict[sheet_name])
                sheet_writer.close()
            else:
                raise Exception("Cannot create a sheet writer!")

    def create_sheet(self, sheet_name):
        """
        implement this method for easy extension
        """
        raise NotImplementedError("Please implement create_sheet()")


def _convert_content_to_stream(file_content, file_type):
    stream = manager.get_io(file_type)
    stream.write(file_content)
    stream.seek(0)
    return stream



"""
    pyexcel_io.database.django
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level handler for django import and export

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import logging

from pyexcel_io.book import BookWriter
from pyexcel_io.sheet import SheetWriter
from pyexcel_io.utils import is_empty_array, swap_empty_string_for_none
import pyexcel_io.constants as constants

log = logging.getLogger(__name__)


class DjangoModelWriter(SheetWriter):
    """ import data into a django model """

    def __init__(self, importer, adapter, batch_size=None, bulk_save=True):
        SheetWriter.__init__(self, importer, adapter, adapter.name)
        self.__batch_size = batch_size
        self.__model = adapter.model
        self.__column_names = adapter.column_names
        self.__mapdict = adapter.column_name_mapping_dict
        self.__initializer = adapter.row_initializer
        self.__objs = []
        self.__bulk_save = bulk_save

    def write_row(self, array):
        if is_empty_array(array):
            print(constants.MESSAGE_EMPTY_ARRAY)
        else:
            new_array = swap_empty_string_for_none(array)
            model_to_be_created = new_array
            if self.__initializer is not None:
                model_to_be_created = self.__initializer(new_array)
            if model_to_be_created:
                self.__objs.append(
                    self.__model(
                        **dict(zip(self.__column_names, model_to_be_created))
                    )
                )

    # else
    # skip the row

    def close(self):
        if self.__bulk_save:
            self.__model.objects.bulk_create(
                self.__objs, batch_size=self.__batch_size
            )
        else:
            for an_object in self.__objs:
                an_object.save()


class DjangoBookWriter(BookWriter):
    """ write data into django models """

    def __init__(self):
        BookWriter.__init__(self)
        self.__importer = None

    def open_content(self, file_content, **keywords):
        self.__importer = file_content
        self._keywords = keywords

    def create_sheet(self, sheet_name):
        sheet_writer = None
        model = self.__importer.get(sheet_name)
        if model:
            sheet_writer = DjangoModelWriter(
                self.__importer,
                model,
                batch_size=self._keywords.get("batch_size", None),
                bulk_save=self._keywords.get("bulk_save", True),
            )
        else:
            raise Exception(
                "Sheet: %s does not match any given models." % sheet_name
                + "Please be aware of case sensitivity."
            )

        return sheet_writer



import os
from nose.tools import eq_
from pyexcel_io import get_data, save_data


class TestRenderer:
    def setUp(self):
        self.test_file = "test_filter.csv"
        sample = [[1, 21, 31], [2, 22, 32]]
        save_data(self.test_file, sample)

    def test_filter_row(self):
        def custom_row_renderer(row):
            return [str(element) for element in row]

        custom_data = get_data(
            self.test_file, row_renderer=custom_row_renderer
        )
        expected = [["1", "21", "31"], ["2", "22", "32"]]
        eq_(custom_data[self.test_file], expected)

    def tearDown(self):
        os.unlink(self.test_file)



"""
    pyexcel_io
    ~~~~~~~~~~~~~~~~~~~

    Uniform interface for reading/writing different excel file formats

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import logging
from ._compact import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())  # noqa

from .io import get_data, iget_data, save_data  # noqa
import pyexcel_io.plugins as plugins


BLACK_LIST = [__name__, "pyexcel_webio", "pyexcel_text"]
WHITE_LIST = [
    "pyexcel_io.readers",
    "pyexcel_io.writers",
    "pyexcel_io.database",
]
PREFIX_PATTERN = "^pyexcel_.*$"

plugins.load_plugins(
    PREFIX_PATTERN,
    __path__,  # noqa: F821
    BLACK_LIST,
    WHITE_LIST)



"""
    pyexcel_io.database.sql
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level handler for database import and export

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.database.common import DbExporter
from pyexcel_io.database.querysets import QuerysetsReader


class SQLTableReader(QuerysetsReader):
    """Read a table
    """

    def __init__(self, session, table, export_columns=None, **keywords):
        everything = session.query(table).all()
        column_names = None
        if export_columns:
            column_names = export_columns
        else:
            if len(everything) > 0:
                column_names = sorted(
                    [
                        column
                        for column in everything[0].__dict__
                        if column != "_sa_instance_state"
                    ]
                )
        QuerysetsReader.__init__(self, everything, column_names, **keywords)


class SQLBookReader(DbExporter):
    """ read a table via sqlalchemy """

    def __init__(self):
        DbExporter.__init__(self)
        self.__exporter = None

    def export_tables(self, file_content, **keywords):
        self.__exporter = file_content
        self._load_from_tables()

    def read_sheet(self, native_sheet):
        reader = SQLTableReader(
            self.__exporter.session,
            native_sheet.table,
            native_sheet.export_columns,
        )
        return reader.to_array()

    def _load_from_tables(self):
        self._native_book = self.__exporter.adapters



"""
    pyexcel_io.writers
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    file writers

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.plugins import IOPluginInfoChain


IOPluginInfoChain(__name__).add_a_writer(
    relative_plugin_class_path="csvw.CSVBookWriter",
    file_types=["csv"],
    stream_type="text",
).add_a_writer(
    relative_plugin_class_path="tsv.TSVBookWriter",
    file_types=["tsv"],
    stream_type="text",
).add_a_writer(
    relative_plugin_class_path="csvz.CSVZipBookWriter",
    file_types=["csvz"],
    stream_type="binary",
).add_a_writer(
    relative_plugin_class_path="tsvz.TSVZipBookWriter",
    file_types=["tsvz"],
    stream_type="binary",
)



"""
    pyexcel_io.fileformat.tsv
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The lower level tsv file format handler.

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import pyexcel_io.constants as constants
from .csvw import CSVBookWriter


class TSVBookWriter(CSVBookWriter):
    """ write tsv """

    def __init__(self):
        CSVBookWriter.__init__(self)
        self._file_type = constants.FILE_FORMAT_TSV

    def open(self, file_name, **keywords):
        keywords["dialect"] = constants.KEYWORD_TSV_DIALECT
        CSVBookWriter.open(self, file_name, **keywords)



import os
import sys
import types
from unittest import TestCase
import pyexcel_io.manager as manager
import pyexcel_io.exceptions as exceptions
from pyexcel_io._compact import StringIO, BytesIO, is_string
from pyexcel_io._compact import OrderedDict
from pyexcel_io import save_data, get_data, iget_data
from pyexcel_io.io import load_data, get_writer
from nose.tools import raises, eq_
from zipfile import BadZipfile


PY2 = sys.version_info[0] == 2


def test_force_file_type():
    test_file = "force_file_type.txt"
    data = get_data(
        os.path.join("tests", "fixtures", test_file), force_file_type="csv"
    )
    expected = [[1, 2, 3]]
    eq_(expected, data[test_file])


@raises(IOError)
def test_no_valid_parameters():
    load_data()


@raises(IOError)
def test_no_valid_parameters_2():
    get_writer()


@raises(IOError)
def test_none_type_load_data():
    get_data(None)


@raises(Exception)
def test_wrong_parameter_to_get_data():
    get_data(1)


@raises(Exception)
def test_wrong_parameter_to_get_writer():
    get_writer(1)


@raises(Exception)
def test_wrong_parameter_to_get_writer2():
    get_writer(1, file_type="csv")


def test_load_ods_data():
    msg = "Please install one of these plugins for read data in 'ods': "
    msg += "pyexcel-ods,pyexcel-ods3"
    try:
        get_data("test.ods")
    except exceptions.SupportingPluginAvailableButNotInstalled as e:
        eq_(str(e), msg)


def test_load_ods_data_from_memory():
    io = BytesIO()
    msg = "Please install one of these plugins for read data in 'ods': "
    msg += "pyexcel-ods,pyexcel-ods3"
    try:
        get_data(io, file_type="ods")
    except exceptions.SupportingPluginAvailableButNotInstalled as e:
        eq_(str(e), msg)


def test_write_xlsx_data_to_memory():
    data = {"Sheet": [[1]]}
    io = BytesIO()
    msg = "Please install one of these plugins for write data in 'xlsx': "
    msg += "pyexcel-xlsx,pyexcel-xlsxw"
    try:
        save_data(io, data, file_type="xlsx")
    except exceptions.SupportingPluginAvailableButNotInstalled as e:
        eq_(str(e), msg)


@raises(exceptions.NoSupportingPluginFound)
def test_load_unknown_data():
    get_data("test.unknown")


@raises(exceptions.NoSupportingPluginFound)
def test_load_unknown_data_from_memory():
    io = BytesIO()
    get_data(io, file_type="unknown")


@raises(BadZipfile)
def test_load_csvz_data_from_memory():
    if not PY2:
        io = StringIO()
        get_data(io, file_type="csvz")
    else:
        raise BadZipfile("pass it")


@raises(IOError)
def test_write_xlsx_data():
    get_data("test.xlsx")


@raises(exceptions.NoSupportingPluginFound)
def test_write_unknown_data():
    get_data("test.unknown")


@raises(Exception)
def test_writer_csvz_data_from_memory():
    if not PY2:
        io = StringIO()
        writer = get_writer(io, file_type="csvz")
        writer.write({"adb": [[2, 3]]})
    else:
        raise Exception("pass it")


@raises(exceptions.NoSupportingPluginFound)
def test_writer_xlsm_data_from_memory2():
    io = BytesIO()
    get_writer(io, file_type="xlsms")


@raises(exceptions.NoSupportingPluginFound)
def test_writer_unknown_data_from_memory2():
    io = BytesIO()
    # mock it
    manager.register_stream_type("unknown1", "text")
    get_writer(io, file_type="unknown1")


def test_get_io():
    io = manager.get_io("hello")
    assert io is None


def test_get_io_type():
    t = manager.get_io_type("hello")
    assert t is None
    t = manager.get_io_type("csv")
    eq_(t, "string")
    t = manager.get_io_type("xls")
    eq_(t, "bytes")


def test_default_csv_format():
    data = [["1", "2", "3"]]
    io = manager.get_io("csv")
    # test default format for saving is 'csv'
    save_data(io, data)
    io.seek(0)
    # test default format for reading is 'csv'
    result = get_data(io)
    assert result["csv"] == [[1, 2, 3]]


def test_case_insentivity():
    data = [["1", "2", "3"]]
    io = manager.get_io("CSV")
    # test default format for saving is 'csv'
    save_data(io, data)
    io.seek(0)
    # test default format for reading is 'csv'
    result = get_data(io)
    assert result["csv"] == [[1, 2, 3]]


def test_file_handle_as_input():
    test_file = "file_handle.csv"
    with open(test_file, "w") as f:
        f.write("1,2,3")

    with open(test_file, "r") as f:
        data = get_data(f, "csv")
        eq_(data["csv"], [[1, 2, 3]])


def test_file_type_case_insensitivity():
    test_file = "file_handle.CSv"
    with open(test_file, "w") as f:
        f.write("1,2,3")

    with open(test_file, "r") as f:
        data = get_data(f, "csv")
        eq_(data["csv"], [[1, 2, 3]])


def test_file_handle_as_output():
    test_file = "file_handle.csv"
    with open(test_file, "w") as f:
        save_data(f, [[1, 2, 3]], "csv", lineterminator="\n")

    with open(test_file, "r") as f:
        content = f.read()
        eq_(content, "1,2,3\n")


def test_binary_file_content():
    data = [["1", "2", "3"]]
    io = manager.get_io("csvz")
    save_data(io, data, "csvz")
    result = get_data(io.getvalue(), "csvz")
    eq_(result["pyexcel_sheet1"], [[1, 2, 3]])


def test_text_file_content():
    data = [["1", "2", "3"]]
    io = manager.get_io("csv")
    save_data(io, data, "csv")
    result = get_data(io.getvalue(), "csv")
    eq_(result["csv"], [[1, 2, 3]])


def test_library_parameter():
    data = [["1", "2", "3"]]
    io = manager.get_io("csv")
    save_data(io, data, "csv", library="pyexcel-io")
    result = get_data(io.getvalue(), "csv", library="pyexcel-io")
    eq_(result["csv"], [[1, 2, 3]])


@raises(Exception)
def test_library_parameter_error_situation():
    data = [["1", "2", "3"]]
    io = manager.get_io("csv")
    save_data(io, data, "csv", library="doesnot-exist")


def test_conversion_from_bytes_to_text():
    test_file = "conversion.csv"
    data = [["1", "2", "3"]]
    save_data(test_file, data)
    with open(test_file, "rb") as f:
        content = f.read()
        result = get_data(content, "csv")
        assert result["csv"] == [[1, 2, 3]]
    os.unlink(test_file)


def test_is_string():
    if PY2:
        assert is_string(type(u"a")) is True
    else:
        assert is_string(type("a")) is True


def test_generator_is_obtained():
    data, reader = iget_data(os.path.join("tests", "fixtures", "test.csv"))
    assert isinstance(data["test.csv"], types.GeneratorType)
    reader.close()


def test_generator_can_be_written():
    test_filename = "generator.csv"
    test_fixture = os.path.join("tests", "fixtures", "test.csv")
    data = get_data(test_fixture, streaming=True)
    save_data(test_filename, data)
    assert os.path.exists(test_filename)
    data2 = get_data(test_filename)
    expected = get_data(test_fixture)
    assert data2[test_filename] == expected["test.csv"]
    os.unlink(test_filename)


class TestReadMultipleSheets(TestCase):
    file_type = "csv"
    delimiter = ","

    def setUp(self):
        self.test_file_formatter = "csv_multiple__%s__%s." + self.file_type
        self.merged_book_file = "csv_multiple." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.expected_data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.sheets = OrderedDict()
        self.sheets.update({"sheet1": self.data})
        self.sheets.update({"sheet2": self.data})
        self.sheets.update({"sheet3": self.data})
        self.expected_sheets = OrderedDict()
        self.expected_sheets.update({"sheet1": self.expected_data})
        self.expected_sheets.update({"sheet2": self.expected_data})
        self.expected_sheets.update({"sheet3": self.expected_data})
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            with open(file_name, "w") as f:
                for row in value:
                    f.write(self.delimiter.join(row) + "\n")
            index = index + 1

    def test_sheet_name(self):
        sheets = get_data(self.merged_book_file, sheet_name="sheet1")
        eq_(sheets["sheet1"], self.expected_sheets["sheet1"])

    def test_sheet_index(self):
        sheets = get_data(self.merged_book_file, sheet_index=1)
        eq_(sheets["sheet2"], self.expected_sheets["sheet2"])

    def test_read_many(self):
        sheets = get_data(self.merged_book_file, sheets=["sheet1", 2])
        eq_(sheets["sheet1"], self.expected_sheets["sheet1"])
        eq_(sheets["sheet3"], self.expected_sheets["sheet3"])
        assert "sheet2" not in sheets

    def tearDown(self):
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            os.unlink(file_name)
            index = index + 1



from pyexcel_io.sheet import SheetReader, SheetWriter, NamedContent
from pyexcel_io.book import BookWriter
from pyexcel_io.utils import is_empty_array
from nose.tools import raises


@raises(NotImplementedError)
def test_book_writer():
    book = BookWriter()
    book.create_sheet("test")


def test_is_empty_array():
    a = ["", "", "", ""]
    assert is_empty_array(a) is True
    b = [1, "", "", ""]
    assert is_empty_array(b) is False


class ArrayWriter(SheetWriter):
    def set_sheet_name(self, name):
        self._native_sheet.name = name

    def write_row(self, array):
        self._native_sheet.payload.append(array)


class TestSheetReader:
    @raises(Exception)
    def test_abstractness(self):
        reader = SheetReader("test")
        reader.cell_value(1, 2)

    @raises(Exception)
    def test_number_of_columns(self):
        reader = SheetReader("test")
        reader.number_of_columns()

    @raises(Exception)
    def test_number_of_rows(self):
        reader = SheetReader("test")
        reader.number_of_rows()

    def test_to_array(self):
        name = "test"

        class B(SheetReader):
            @property
            def name(self):
                return self._native_sheet

            def to_array(self):
                pass

        b = B(name)
        b.to_array()
        assert b.name == name


class TestSheetWriter:
    @raises(NotImplementedError)
    def test_abstractness(self):
        writer = SheetWriter("te", "st", "abstract")
        writer.write_row([])

    def test_inheritance(self):
        class D(SheetWriter):
            def write_row(self, row):
                pass

        d = D("t", "e", "s")
        d.write_row([11, 11])

    def test_writer(self):
        native_sheet = NamedContent("test", [])
        content = [[1, 2], [3, 4], [5, 6]]
        writer = ArrayWriter(None, native_sheet, "test")
        writer.write_row(content[0])
        writer.write_array(content[1:])
        assert native_sheet.payload == content

    def test_writer2(self):
        native_sheet = NamedContent("test", [])
        content = [[1, 2], [3, 4], [5, 6]]
        writer = ArrayWriter(None, native_sheet, None)
        writer.write_row(content[0])
        writer.write_array(content[1:])
        assert native_sheet.payload == content
        assert native_sheet.name == "pyexcel_sheet1"



import os
from unittest import TestCase
from textwrap import dedent
from nose.tools import raises
import pyexcel_io.manager as manager
from pyexcel_io._compact import OrderedDict
from pyexcel_io.readers.csvr import CSVBookReader
from pyexcel_io.readers.tsv import TSVBookReader
from pyexcel_io.writers.csvw import CSVBookWriter
from pyexcel_io.writers.tsv import TSVBookWriter


class TestCSVReaders(TestCase):
    file_type = "csv"
    reader_class = CSVBookReader
    delimiter = ","

    def setUp(self):
        self.test_file = "csv_book." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.expected_data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        with open(self.test_file, "w") as f:
            for row in self.data:
                f.write(self.delimiter.join(row) + "\n")

    def test_book_reader(self):
        b = self.reader_class()
        b.open(self.test_file)
        sheets = b.read_all()
        self.assertEqual(list(sheets[self.test_file]), self.expected_data)

    def test_book_reader_from_memory_source(self):
        io = manager.get_io(self.file_type)
        with open(self.test_file, "r") as f:
            io.write(f.read())
        io.seek(0)
        b = self.reader_class()
        b.open_stream(io)
        sheets = b.read_all()
        self.assertEqual(list(sheets[self.file_type]), self.expected_data)

    def tearDown(self):
        os.unlink(self.test_file)


class TestTSVReaders(TestCSVReaders):
    file_type = "tsv"
    reader_class = TSVBookReader
    delimiter = "\t"


class TestReadMultipleSheets(TestCase):
    file_type = "csv"
    reader_class = CSVBookReader
    delimiter = ","

    def setUp(self):
        self.test_file_formatter = "csv_multiple__%s__%s." + self.file_type
        self.merged_book_file = "csv_multiple." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.expected_data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.sheets = OrderedDict()
        self.sheets.update({"sheet1": self.data})
        self.sheets.update({"sheet2": self.data})
        self.sheets.update({"sheet3": self.data})
        self.expected_sheets = OrderedDict()
        self.expected_sheets.update({"sheet1": self.expected_data})
        self.expected_sheets.update({"sheet2": self.expected_data})
        self.expected_sheets.update({"sheet3": self.expected_data})
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            with open(file_name, "w") as f:
                for row in value:
                    f.write(self.delimiter.join(row) + "\n")
            index = index + 1

    def test_multiple_sheet(self):
        b = self.reader_class()
        b.open(self.merged_book_file)
        sheets = b.read_all()
        for key in sheets:
            sheets[key] = list(sheets[key])
        self.assertEqual(sheets, self.expected_sheets)

    def test_read_one_from_many_by_name(self):
        b = self.reader_class()
        b.open(self.merged_book_file)
        sheets = b.read_sheet_by_name("sheet1")
        self.assertEqual(
            list(sheets["sheet1"]), self.expected_sheets["sheet1"]
        )

    @raises(ValueError)
    def test_read_one_from_many_by_non_existent_name(self):
        b = self.reader_class()
        b.open(self.merged_book_file)
        b.read_sheet_by_name("notknown")

    def test_read_one_from_many_by_index(self):
        b = self.reader_class()
        b.open(self.merged_book_file)
        sheets = b.read_sheet_by_index(1)
        self.assertEqual(
            list(sheets["sheet2"]), self.expected_sheets["sheet2"]
        )

    @raises(IndexError)
    def test_read_one_from_many_by_wrong_index(self):
        b = self.reader_class()
        b.open(self.merged_book_file)
        b.read_sheet_by_index(90)

    def tearDown(self):
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            os.unlink(file_name)
            index = index + 1


class TestTSVBookReaders(TestReadMultipleSheets):
    file_type = "tsv"
    reader_class = TSVBookReader
    delimiter = "\t"


class TestWriteMultipleSheets(TestCase):
    file_type = "csv"
    writer_class = CSVBookWriter
    reader_class = CSVBookReader
    result1 = dedent(
        """
        1,2,3
        4,5,6
        7,8,9
        """
    ).strip("\n")
    result2 = dedent(
        """
        1,2,3
        4,5,6
        7,8,1000
        """
    ).strip("\n")
    result3 = dedent(
        """
        1,2,3
        4,5,6888
        7,8,9
        """
    ).strip("\n")
    merged = dedent(
        """\
        ---pyexcel:sheet1---
        1,2,3
        4,5,6
        7,8,9
        ---pyexcel---
        ---pyexcel:sheet2---
        1,2,3
        4,5,6
        7,8,1000
        ---pyexcel---
        ---pyexcel:sheet3---
        1,2,3
        4,5,6888
        7,8,9
        ---pyexcel---
        """
    )

    def setUp(self):
        self.test_file_formatter = "csv_multiple__%s__%s." + self.file_type
        self.merged_book_file = "csv_multiple." + self.file_type
        self.data1 = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]
        self.data2 = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "1000"]]
        self.data3 = [["1", "2", "3"], ["4", "5", "6888"], ["7", "8", "9"]]
        self.expected_data1 = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        self.expected_data2 = [[1, 2, 3], [4, 5, 6], [7, 8, 1000]]
        self.expected_data3 = [[1, 2, 3], [4, 5, 6888], [7, 8, 9]]
        self.sheets = OrderedDict()
        self.sheets.update({"sheet1": self.data1})
        self.sheets.update({"sheet2": self.data2})
        self.sheets.update({"sheet3": self.data3})
        self.expected_sheets = OrderedDict()
        self.expected_sheets.update({"sheet1": self.expected_data1})
        self.expected_sheets.update({"sheet2": self.expected_data2})
        self.expected_sheets.update({"sheet3": self.expected_data3})
        self.result_dict = OrderedDict()
        self.result_dict.update({"sheet1": self.result1})
        self.result_dict.update({"sheet2": self.result2})
        self.result_dict.update({"sheet3": self.result3})

    def test_multiple_sheet(self):
        w = self.writer_class()
        w.open(self.merged_book_file)
        w.write(self.sheets)
        w.close()
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            with open(file_name, "r") as f:
                content = f.read().replace("\r", "")
                assert content.strip("\n") == self.result_dict[key]
            index = index + 1
        self.delete_files()

    def test_multiple_sheet_into_memory(self):
        io = manager.get_io(self.file_type)
        w = self.writer_class()
        w.open(io, lineterminator="\n")
        w.write(self.sheets)
        w.close()
        content = io.getvalue()
        self.assertEqual(content, self.merged)

    def test_multiple_sheet_into_memory_2(self):
        """Write csv book into a single stream"""
        io = manager.get_io(self.file_type)
        w = self.writer_class()
        w.open(io, lineterminator="\n")
        w.write(self.sheets)
        w.close()
        reader = self.reader_class()
        reader.open_stream(io, lineterminator="\n", multiple_sheets=True)
        sheets = reader.read_all()
        for sheet in sheets:
            sheets[sheet] = list(sheets[sheet])
        self.assertEqual(sheets, self.expected_sheets)

    def delete_files(self):
        index = 0
        for key, value in self.sheets.items():
            file_name = self.test_file_formatter % (key, index)
            os.unlink(file_name)
            index = index + 1


class TestTSVWriteMultipleSheets(TestWriteMultipleSheets):
    file_type = "tsv"
    writer_class = TSVBookWriter
    reader_class = TSVBookReader
    result1 = dedent(
        """
        1\t2\t3
        4\t5\t6
        7\t8\t9
        """
    ).strip("\n")
    result2 = dedent(
        """
        1\t2\t3
        4\t5\t6
        7\t8\t1000
        """
    ).strip("\n")
    result3 = dedent(
        """
        1\t2\t3
        4\t5\t6888
        7\t8\t9
        """
    ).strip("\n")
    merged = dedent(
        """\
        ---pyexcel:sheet1---
        1\t2\t3
        4\t5\t6
        7\t8\t9
        ---pyexcel---
        ---pyexcel:sheet2---
        1\t2\t3
        4\t5\t6
        7\t8\t1000
        ---pyexcel---
        ---pyexcel:sheet3---
        1\t2\t3
        4\t5\t6888
        7\t8\t9
        ---pyexcel---
        """
    )


class TestWriter(TestCase):
    file_type = "csv"
    writer_class = CSVBookWriter
    result = dedent(
        """
        1,2,3
        4,5,6
        7,8,9
        """
    ).strip("\n")

    def setUp(self):
        self.test_file = "csv_book." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]

    def test_book_writer(self):
        w = self.writer_class()
        w.open(self.test_file)
        w.write({None: self.data})
        w.close()
        with open(self.test_file, "r") as f:
            content = f.read().replace("\r", "")
            self.assertEqual(content.strip("\n"), self.result)

    def tearDown(self):
        os.unlink(self.test_file)


class TestTSVWriters(TestWriter):
    file_type = "tsv"
    writer_class = TSVBookWriter
    result = dedent(
        """
        1\t2\t3
        4\t5\t6
        7\t8\t9
        """
    ).strip("\n")


class TestMemoryWriter(TestCase):
    file_type = "csv"
    writer_class = CSVBookWriter
    result = dedent(
        """
           1,2,3
           4,5,6
           7,8,9
        """
    ).strip("\n")

    def setUp(self):
        self.test_file = "csv_book." + self.file_type
        self.data = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"]]

    def test_book_writer_to_memroy(self):
        io = manager.get_io(self.file_type)
        w = self.writer_class()
        w.open(io, single_sheet_in_book=True)
        w.write({self.file_type: self.data})
        w.close()
        content = io.getvalue().replace("\r", "")
        assert content.strip("\n") == self.result


class TestTSVMemoryWriter(TestMemoryWriter):
    file_type = "tsv"
    writer_class = TSVBookWriter
    result = dedent(
        """
           1\t2\t3
           4\t5\t6
           7\t8\t9
        """
    ).strip("\n")



# -*- coding: utf-8 -*-
import os
from unittest import TestCase
from pyexcel_io._compact import OrderedDict
from pyexcel_io import save_data
import pyexcel_io.manager as manager
from pyexcel_io.readers.csvz import CSVZipBookReader
from pyexcel_io.writers.csvz import CSVZipBookWriter
from pyexcel_io.readers.tsvz import TSVZipBookReader
from pyexcel_io.writers.tsvz import TSVZipBookWriter
import zipfile
from nose.tools import raises
import sys

PY2 = sys.version_info[0] == 2


class TestCSVZ(TestCase):
    file_type = "csvz"
    writer_class = CSVZipBookWriter
    reader_class = CSVZipBookReader
    result = u"中,文,1,2,3"

    def setUp(self):
        self.file = "csvz." + self.file_type

    def test_writing(self):
        data = [[u"中", u"文", 1, 2, 3]]
        file_name = "pyexcel_sheet1." + self.file_type[0:3]
        zipbook = self.writer_class()
        zipbook.open(self.file)
        zipbook.write({None: data})
        zipbook.close()
        zip = zipfile.ZipFile(self.file, "r")
        self.assertEqual(zip.namelist(), [file_name])
        content = zip.read(file_name)
        content = content.decode("utf-8")
        self.assertEqual(content.replace("\r", "").strip("\n"), self.result)
        zip.close()

    def test_reading(self):
        data = [[u"中", u"文", 1, 2, 3]]
        zipbook = self.writer_class()
        zipbook.open(self.file)
        zipbook.write({None: data})
        zipbook.close()
        zipreader = self.reader_class()
        zipreader.open(self.file)
        data = zipreader.read_all()
        self.assertEqual(list(data["pyexcel_sheet1"]), [[u"中", u"文", 1, 2, 3]])
        zipreader.close()

    def tearDown(self):
        os.unlink(self.file)


class TestTSVZ(TestCSVZ):
    file_type = "tsvz"
    writer_class = TSVZipBookWriter
    reader_class = TSVZipBookReader
    result = u"中\t文\t1\t2\t3"


def test_reading_from_memory():
    data = [[1, 2, 3]]
    io = manager.get_io("csvz")
    zipbook = CSVZipBookWriter()
    zipbook.open_stream(io)
    zipbook.write({None: data})
    zipbook.close()
    zipreader = CSVZipBookReader()
    zipreader.open_stream(io)
    data = zipreader.read_all()
    assert list(data["pyexcel_sheet1"]) == [[1, 2, 3]]


def test_reading_from_memory_tsvz():
    data = [[1, 2, 3]]
    io = manager.get_io("tsvz")
    zipbook = TSVZipBookWriter()
    zipbook.open_stream(io)
    zipbook.write({None: data})
    zipbook.close()
    zipreader = TSVZipBookReader()
    zipreader.open_stream(io)
    data = zipreader.read_all()
    assert list(data["pyexcel_sheet1"]) == [[1, 2, 3]]


class TestMultipleSheet(TestCase):
    file_name = "mybook.csvz"
    reader_class = CSVZipBookReader

    def setUp(self):
        self.content = OrderedDict()
        self.content.update(
            {"Sheet 1": [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]}
        )
        self.content.update(
            {"Sheet 2": [["X", "Y", "Z"], [1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]}
        )
        self.content.update(
            {"Sheet 3": [["O", "P", "Q"], [3.0, 2.0, 1.0], [4.0, 3.0, 2.0]]}
        )
        save_data(self.file_name, self.content)

    def test_read_one_from_many_by_name(self):
        reader = self.reader_class()
        reader.open(self.file_name)
        sheets = reader.read_sheet_by_name("Sheet 1")
        self.assertEqual(
            list(sheets["Sheet 1"]), [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )

    @raises(ValueError)
    def test_read_one_from_many_by_unknown_name(self):
        reader = self.reader_class()
        reader.open(self.file_name)
        reader.read_sheet_by_name("Sheet X")

    def test_read_one_from_many_by_index(self):
        reader = self.reader_class()
        reader.open(self.file_name)
        sheets = reader.read_sheet_by_index(0)
        self.assertEqual(
            list(sheets["Sheet 1"]), [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )

    @raises(IndexError)
    def test_read_one_from_many_by_unknown_index(self):
        reader = self.reader_class()
        reader.open(self.file_name)
        reader.read_sheet_by_index(9999)

    def tearDown(self):
        os.unlink(self.file_name)


class TestMultipleTSVSheet(TestMultipleSheet):
    file_name = "mybook.tsvz"
    reader_class = TSVZipBookReader



"""
    pyexcel_io.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~

    all possible exceptions

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""


class NoSupportingPluginFound(Exception):
    """raised when an known file extension is seen"""

    pass


class SupportingPluginAvailableButNotInstalled(Exception):
    """raised when a known plugin is not installed"""

    pass


class UpgradePlugin(Exception):
    """raised when a known plugin is not compatible"""

    pass


class IntegerAccuracyLossError(Exception):
    """
    When an interger is greater than 999999999999999, ods loses its accuracy.

    from pyexcel import Sheet, get_sheet
    s = Sheet()
    s[0,0] = 999999999999999  # 15 '9's
    print(s)
    s.save_as('abc.ods')
    b=get_sheet(file_name='abc.ods')
    b[0,0] == s[0,0]

    s = Sheet()
    s[0,0] = 9999999999999999 # 16 '9's
    print(s)
    s.save_as('abc.ods')
    b=get_sheet(file_name='abc.ods')
    b[0,0] != s[0,0]
    """
    def __init__(self, message):
        custom_message = (
            message + '\n' +
            "In order to keep its accuracy, please save as string. Then " +
            "convert to int, long or float after the value will be read back"
        )

        super(IntegerAccuracyLossError, self).__init__(custom_message)



"""
    pyexcel_io.readers.csvr
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    csv file reader

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
import re
import os
import csv
import glob
import codecs

from pyexcel_io.book import BookReader
from pyexcel_io.sheet import SheetReader, NamedContent
import pyexcel_io._compact as compact
import pyexcel_io.constants as constants
import pyexcel_io.service as service


DEFAULT_SEPARATOR = "__"
DEFAULT_SHEET_SEPARATOR_FORMATTER = "---%s---" % constants.DEFAULT_NAME + "%s"
SEPARATOR_MATCHER = "---%s:(.*)---" % constants.DEFAULT_NAME
DEFAULT_CSV_STREAM_FILE_FORMATTER = (
    "---%s:" % constants.DEFAULT_NAME + "%s---%s"
)
DEFAULT_NEWLINE = "\r\n"
BOM_LITTLE_ENDIAN = b"\xff\xfe"
BOM_BIG_ENDIAN = b"\xfe\ff"
LITTLE_ENDIAN = 0
BIG_ENDIAN = 1


class CSVMemoryMapIterator(compact.Iterator):
    """
    Wrapper class for mmap object

    mmap object does not handle encoding at all. This class
    provide the necessary transcoding for utf-8, utf-16 and utf-32
    """

    def __init__(self, mmap_obj, encoding):
        self.__mmap_obj = mmap_obj
        self.__encoding = encoding
        self.__count = 0
        self.__endian = LITTLE_ENDIAN
        if encoding == "utf-8":
            # ..\r\x00\n
            # \x00\x..
            self.__zeros_left_in_2_row = 0
        elif encoding == "utf-16":
            # ..\r\x00\n
            # \x00\x..
            self.__zeros_left_in_2_row = 1
        elif encoding == "utf-32":
            # \r\x00\x00\x00\n
            # \x00\x00\x00\x..
            self.__zeros_left_in_2_row = 3
        elif encoding == "utf-32-be" or encoding == "utf-16-be":
            self.__zeros_left_in_2_row = 0
            self.__endian = BIG_ENDIAN
        elif encoding == "utf-32-le":
            self.__zeros_left_in_2_row = 3
            self.__endian = LITTLE_ENDIAN
        elif encoding == "utf-16-le":
            self.__zeros_left_in_2_row = 1
            self.__endian = LITTLE_ENDIAN
        else:
            raise Exception("Encoding %s is not supported" % encoding)

    def __iter__(self):
        return self

    def __next__(self):
        line = self.__mmap_obj.readline()
        if self.__count == 0:
            utf_16_32 = (
                self.__encoding == "utf-16" or self.__encoding == "utf-32"
            )
            if utf_16_32:
                bom_header = line[:2]
                if bom_header == BOM_BIG_ENDIAN:
                    self.__endian = BIG_ENDIAN
        elif self.__endian == LITTLE_ENDIAN:
            line = line[self.__zeros_left_in_2_row :]  # noqa: E203
        if self.__endian == LITTLE_ENDIAN:
            line = line.rstrip()
        line = line.decode(self.__encoding)
        self.__count += 1
        if line == "":
            raise StopIteration

        if compact.PY2:
            # python 2 requires utf-8 encoded string for reading
            line = line.encode("utf-8")
        return line


class UTF8Recorder(compact.Iterator):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8.
    """

    def __init__(self, file_handle, encoding):
        self.__file_handle = file_handle
        self.reader = codecs.getreader(encoding)(file_handle)

    def close(self):
        self.__file_handle.close()

    def __iter__(self):
        return self

    def __next__(self):
        # python 2 requires utf-8 encoded string for reading
        line = next(self.reader).encode("utf-8")
        return line


class CSVSheetReader(SheetReader):
    """ generic csv file reader"""

    def __init__(
        self,
        sheet,
        encoding="utf-8",
        auto_detect_float=True,
        ignore_infinity=True,
        auto_detect_int=True,
        auto_detect_datetime=True,
        pep_0515_off=True,
        ignore_nan_text=False,
        default_float_nan=None,
        **keywords
    ):
        SheetReader.__init__(self, sheet, **keywords)
        self._encoding = encoding
        self.__auto_detect_int = auto_detect_int
        self.__auto_detect_float = auto_detect_float
        self.__ignore_infinity = ignore_infinity
        self.__auto_detect_datetime = auto_detect_datetime
        self.__file_handle = None
        self.__pep_0515_off = pep_0515_off
        self.__ignore_nan_text = ignore_nan_text
        self.__default_float_nan = default_float_nan

    def get_file_handle(self):
        """ return me unicde reader for csv """
        raise NotImplementedError("Please implement get_file_handle()")

    def row_iterator(self):
        self.__file_handle = self.get_file_handle()
        return csv.reader(self.__file_handle, **self._keywords)

    def column_iterator(self, row):
        for element in row:
            if compact.PY2:
                element = element.decode("utf-8")
            if element is not None and element != "":
                element = self.__convert_cell(element)
            yield element

    def __convert_cell(self, csv_cell_text):
        ret = None
        if self.__auto_detect_int:
            ret = service.detect_int_value(csv_cell_text, self.__pep_0515_off)
        if ret is None and self.__auto_detect_float:
            ret = service.detect_float_value(
                csv_cell_text,
                self.__pep_0515_off,
                ignore_nan_text=self.__ignore_nan_text,
                default_float_nan=self.__default_float_nan,
            )
            shall_we_ignore_the_conversion = (
                ret in [float("inf"), float("-inf")]
            ) and self.__ignore_infinity
            if shall_we_ignore_the_conversion:
                ret = None
        if ret is None and self.__auto_detect_datetime:
            ret = service.detect_date_value(csv_cell_text)
        if ret is None:
            ret = csv_cell_text
        return ret

    def close(self):
        if self.__file_handle:
            self.__file_handle.close()


# else: means the generator has been run
# yes, no run, no file open.


class CSVFileReader(CSVSheetReader):
    """ read csv from phyical file """

    def get_file_handle(self):
        unicode_reader = None
        if compact.PY2:
            file_handle = open(self._native_sheet.payload, "rb")
            unicode_reader = UTF8Recorder(file_handle, self._encoding)
        else:
            unicode_reader = open(
                self._native_sheet.payload, "r", encoding=self._encoding
            )
        return unicode_reader


class CSVinMemoryReader(CSVSheetReader):
    """ read csv file from memory """

    def get_file_handle(self):
        unicode_reader = None
        if compact.PY2:
            if hasattr(self._native_sheet.payload, "read"):
                unicode_reader = UTF8Recorder(
                    self._native_sheet.payload, self._encoding
                )
            else:
                unicode_reader = self._native_sheet.payload
        else:
            if isinstance(self._native_sheet.payload, compact.BytesIO):
                # please note that
                # if the end developer feed us bytesio in python3
                # we will do the conversion to StriongIO but that
                # comes at a cost.
                content = self._native_sheet.payload.read()
                unicode_reader = compact.StringIO(
                    content.decode(self._encoding)
                )
            else:
                unicode_reader = self._native_sheet.payload

        return unicode_reader


class CSVBookReader(BookReader):
    """ read csv file """

    def __init__(self):
        BookReader.__init__(self)
        self._file_type = constants.FILE_FORMAT_CSV
        self._file_content = None
        self.__load_from_memory_flag = False
        self.__line_terminator = constants.DEFAULT_CSV_NEWLINE
        self.__sheet_name = None
        self.__sheet_index = None
        self.__multiple_sheets = False
        self.__readers = []

    def open(self, file_name, **keywords):
        BookReader.open(self, file_name, **keywords)
        self._native_book = self._load_from_file()

    def open_stream(self, file_stream, multiple_sheets=False, **keywords):
        BookReader.open_stream(self, file_stream, **keywords)
        self.__multiple_sheets = multiple_sheets
        self._native_book = self._load_from_stream()

    def open_content(self, file_content, **keywords):
        try:
            import mmap

            encoding = keywords.get("encoding", "utf-8")
            if isinstance(file_content, mmap.mmap):
                # load from mmap
                self.__multiple_sheets = keywords.get("multiple_sheets", False)
                self._file_stream = CSVMemoryMapIterator(
                    file_content, encoding
                )
                self._keywords = keywords
                self._native_book = self._load_from_stream()
            else:
                if compact.PY3_ABOVE:
                    if isinstance(file_content, bytes):
                        file_content = file_content.decode(encoding)
                # else python 2.7 does not care about bytes nor str
                BookReader.open_content(self, file_content, **keywords)
        except ImportError:
            # python 2.6 or Google app engine
            BookReader.open_content(self, file_content, **keywords)

    def read_sheet(self, native_sheet):
        if self.__load_from_memory_flag:
            reader = CSVinMemoryReader(native_sheet, **self._keywords)
        else:
            reader = CSVFileReader(native_sheet, **self._keywords)
            self.__readers.append(reader)
        return reader.to_array()

    def close(self):
        for reader in self.__readers:
            reader.close()

    def _load_from_stream(self):
        """Load content from memory

        :params stream file_content: the actual file content in memory
        :returns: a book
        """
        self.__load_from_memory_flag = True
        self.__line_terminator = self._keywords.get(
            constants.KEYWORD_LINE_TERMINATOR, self.__line_terminator
        )
        separator = DEFAULT_SHEET_SEPARATOR_FORMATTER % self.__line_terminator
        if self.__multiple_sheets:
            # will be slow for large files
            self._file_stream.seek(0)
            content = self._file_stream.read()
            sheets = content.split(separator)
            named_contents = []
            for sheet in sheets:
                if sheet == "":  # skip empty named sheet
                    continue

                lines = sheet.split(self.__line_terminator)
                result = re.match(constants.SEPARATOR_MATCHER, lines[0])
                new_content = "\n".join(lines[1:])
                new_sheet = NamedContent(
                    result.group(1), compact.StringIO(new_content)
                )
                named_contents.append(new_sheet)
            return named_contents

        else:
            if hasattr(self._file_stream, "seek"):
                self._file_stream.seek(0)
            return [NamedContent(self._file_type, self._file_stream)]

    def _load_from_file(self):
        """Load content from a file

        :params str filename: an accessible file path
        :returns: a book
        """
        self.__line_terminator = self._keywords.get(
            constants.KEYWORD_LINE_TERMINATOR, self.__line_terminator
        )
        names = os.path.splitext(self._file_name)
        filepattern = "%s%s*%s*%s" % (
            names[0],
            constants.DEFAULT_MULTI_CSV_SEPARATOR,
            constants.DEFAULT_MULTI_CSV_SEPARATOR,
            names[1],
        )
        filelist = glob.glob(filepattern)
        if len(filelist) == 0:
            file_parts = os.path.split(self._file_name)
            return [NamedContent(file_parts[-1], self._file_name)]

        else:
            matcher = "%s%s(.*)%s(.*)%s" % (
                names[0],
                constants.DEFAULT_MULTI_CSV_SEPARATOR,
                constants.DEFAULT_MULTI_CSV_SEPARATOR,
                names[1],
            )
            tmp_file_list = []
            for filen in filelist:
                result = re.match(matcher, filen)
                tmp_file_list.append((result.group(1), result.group(2), filen))
            ret = []
            for lsheetname, index, filen in sorted(
                tmp_file_list, key=lambda row: row[1]
            ):
                ret.append(NamedContent(lsheetname, filen))
            return ret



from nose.tools import raises, eq_
from pyexcel_io import save_data
from pyexcel_io._compact import OrderedDict
from pyexcel_io.constants import DB_DJANGO
from pyexcel_io.database.common import (
    DjangoModelImporter,
    DjangoModelImportAdapter,
    DjangoModelExporter,
    DjangoModelExportAdapter,
)
from pyexcel_io.database.importers.django import (
    DjangoModelWriter,
    DjangoBookWriter,
)
from pyexcel_io.database.exporters.django import (
    DjangoModelReader,
    DjangoBookReader,
)


class Package:
    def __init__(self, raiseException=False, **keywords):
        self.keywords = keywords
        self.raiseException = raiseException

    def get_content(self):
        return self.keywords

    def save(self):
        if self.raiseException:
            raise Exception("test exception")
        else:
            pass


class Attributable:
    def __init__(self, adict):
        self.mydict = adict

    def __getattr__(self, field):
        return self.mydict[field]


class Objects:
    def __init__(self):
        self.objs = []

    def bulk_create(self, objs, batch_size):
        self.objs = [o.get_content() for o in objs]
        self.batch_size = batch_size

    def all(self):
        return [Attributable(o) for o in self.objs]


class Field:
    def __init__(self, name):
        self.attname = name


class Meta:
    instance = 1

    def __init__(self):
        self.model_name = "Sheet%d" % Meta.instance
        self.concrete_fields = []
        Meta.instance = Meta.instance + 1

    def update(self, data):
        for f in data:
            self.concrete_fields.append(Field(f))


class FakeDjangoModel:
    def __init__(self):
        self.objects = Objects()
        self._meta = Meta()

    def __call__(self, **keywords):
        return Package(**keywords)


class ExceptionObjects(Objects):
    def bulk_create(self, objs, batch_size):
        raise Exception("faked exception")


class FakeExceptionDjangoModel(FakeDjangoModel):
    def __init__(self, raiseException=False):
        self.objects = ExceptionObjects()
        self._meta = Meta()
        self.raiseException = raiseException

    def __call__(self, **keywords):
        return Package(raiseException=self.raiseException, **keywords)


class TestException:
    def setUp(self):
        self.data = [["X", "Y", "Z"], [1, 2, 3], [4, 5, 6]]
        self.result = [{"Y": 2, "X": 1, "Z": 3}, {"Y": 5, "X": 4, "Z": 6}]

    @raises(Exception)
    def test_bulk_save_to_django_model_with_exception(self):
        model = FakeExceptionDjangoModel()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(self.data[1:])
        writer.close()

    @raises(Exception)
    def test_single_save_to_django_model_with_exception(self):
        model = FakeExceptionDjangoModel(raiseException=True)
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(self.data[1:])
        writer.close()


class TestSheet:
    def setUp(self):
        self.data = [["X", "Y", "Z"], [1, 2, 3], [4, 5, 6]]
        self.result = [{"Y": 2, "X": 1, "Z": 3}, {"Y": 5, "X": 4, "Z": 6}]

    def test_sheet_save_to_django_model(self):
        model = FakeDjangoModel()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(self.data[1:])
        writer.close()
        eq_(model.objects.objs, self.result)

    def test_sheet_save_to_django_model_with_empty_array(self):
        model = FakeDjangoModel()
        data = [["X", "Y", "Z"], ["", "", ""], [1, 2, 3], [4, 5, 6]]
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(data[1:])
        writer.close()
        assert model.objects.objs == self.result

    def test_sheet_save_to_django_model_3(self):
        model = FakeDjangoModel()

        def wrapper(row):
            row[0] = row[0] + 1
            return row

        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        adapter.row_initializer = wrapper
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(self.data[1:])
        writer.close()
        assert model.objects.objs == [
            {"Y": 2, "X": 2, "Z": 3},
            {"Y": 5, "X": 5, "Z": 6},
        ]

    def test_sheet_save_to_django_model_skip_me(self):
        model = FakeDjangoModel()

        def wrapper(row):
            if row[0] == 4:
                return None
            else:
                return row

        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        adapter.row_initializer = wrapper
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(self.data[1:])
        writer.close()
        assert model.objects.objs == [{"Y": 2, "X": 1, "Z": 3}]

    def test_load_sheet_from_django_model(self):
        model = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        importer.append(adapter)
        save_data(
            importer, {adapter.get_name(): self.data[1:]}, file_type=DB_DJANGO
        )
        assert model.objects.objs == self.result
        model._meta.update(["X", "Y", "Z"])
        reader = DjangoModelReader(model)
        data = reader.to_array()
        assert list(data) == self.data

    def test_load_and_format_sheet_from_django_model(self):
        model = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = self.data[0]
        importer.append(adapter)
        save_data(
            importer, {adapter.get_name(): self.data[1:]}, file_type=DB_DJANGO
        )
        assert model.objects.objs == self.result
        model._meta.update(["X", "Y", "Z"])

        def row_renderer(row):
            return [str(element) for element in row]

        # the key point of this test case
        reader = DjangoModelReader(model, row_renderer=row_renderer)
        data = reader.to_array()
        expected = [["X", "Y", "Z"], ["1", "2", "3"], ["4", "5", "6"]]
        eq_(list(data), expected)

    def test_mapping_array(self):
        data2 = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        model = FakeDjangoModel()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = data2[0]
        adapter.column_name_mapping_dict = ["X", "Y", "Z"]
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(data2[1:])
        writer.close()
        eq_(model.objects.objs, self.result)

    def test_mapping_dict(self):
        data2 = [["A", "B", "C"], [1, 2, 3], [4, 5, 6]]
        mapdict = {"C": "Z", "A": "X", "B": "Y"}
        model = FakeDjangoModel()
        adapter = DjangoModelImportAdapter(model)
        adapter.column_names = data2[0]
        adapter.column_name_mapping_dict = mapdict
        writer = DjangoModelWriter(None, adapter)
        writer.write_array(data2[1:])
        writer.close()
        eq_(model.objects.objs, self.result)

    def test_empty_model(self):
        model = FakeDjangoModel()
        reader = DjangoModelReader(model)
        data = reader.to_array()
        eq_(list(data), [[], []])


class TestMultipleModels:
    def setUp(self):
        self.content = OrderedDict()
        self.content.update(
            {"Sheet1": [[u"X", u"Y", u"Z"], [1, 4, 7], [2, 5, 8], [3, 6, 9]]}
        )
        self.content.update(
            {"Sheet2": [[u"A", u"B", u"C"], [1, 4, 7], [2, 5, 8], [3, 6, 9]]}
        )
        self.result1 = [
            {"Y": 4, "X": 1, "Z": 7},
            {"Y": 5, "X": 2, "Z": 8},
            {"Y": 6, "X": 3, "Z": 9},
        ]
        self.result2 = [
            {"B": 4, "A": 1, "C": 7},
            {"B": 5, "A": 2, "C": 8},
            {"B": 6, "A": 3, "C": 9},
        ]

    def test_save_to_more_models(self):
        sample_size = 10
        model1 = FakeDjangoModel()
        model2 = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter1 = DjangoModelImportAdapter(model1)
        adapter1.column_names = self.content["Sheet1"][0]
        adapter2 = DjangoModelImportAdapter(model2)
        adapter2.column_names = self.content["Sheet2"][0]
        importer.append(adapter1)
        importer.append(adapter2)
        to_store = {
            adapter1.get_name(): self.content["Sheet1"][1:],
            adapter2.get_name(): self.content["Sheet2"][1:],
        }
        writer = DjangoBookWriter()
        writer.open_content(importer, batch_size=sample_size)
        writer.write(to_store)
        writer.close()
        assert model1.objects.objs == self.result1
        assert model2.objects.objs == self.result2
        assert model1.objects.batch_size == sample_size

    def test_bulk_save_flag(self):
        sample_size = 10
        model1 = FakeDjangoModel()
        model2 = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter1 = DjangoModelImportAdapter(model1)
        adapter1.column_names = self.content["Sheet1"][0]
        adapter2 = DjangoModelImportAdapter(model2)
        adapter2.column_names = self.content["Sheet2"][0]
        importer.append(adapter1)
        importer.append(adapter2)
        to_store = {
            adapter1.get_name(): self.content["Sheet1"][1:],
            adapter2.get_name(): self.content["Sheet2"][1:],
        }
        writer = DjangoBookWriter()
        writer.open_content(importer, batch_size=sample_size, bulk_save=False)
        writer.write(to_store)
        writer.close()
        assert model1.objects.objs == []

    def test_reading_from_more_models(self):
        model1 = FakeDjangoModel()
        model2 = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter1 = DjangoModelImportAdapter(model1)
        adapter1.column_names = self.content["Sheet1"][0]
        adapter2 = DjangoModelImportAdapter(model2)
        adapter2.column_names = self.content["Sheet2"][0]
        importer.append(adapter1)
        importer.append(adapter2)
        to_store = {
            adapter1.get_name(): self.content["Sheet1"][1:],
            adapter2.get_name(): self.content["Sheet2"][1:],
        }
        save_data(importer, to_store, file_type=DB_DJANGO)
        assert model1.objects.objs == self.result1
        assert model2.objects.objs == self.result2
        model1._meta.model_name = "Sheet1"
        model2._meta.model_name = "Sheet2"
        model1._meta.update(["X", "Y", "Z"])
        model2._meta.update(["A", "B", "C"])
        exporter = DjangoModelExporter()
        adapter1 = DjangoModelExportAdapter(model1)
        adapter2 = DjangoModelExportAdapter(model2)
        exporter.append(adapter1)
        exporter.append(adapter2)
        reader = DjangoBookReader()
        reader.open_content(exporter)
        data = reader.read_all()
        for key in data.keys():
            data[key] = list(data[key])
        assert data == self.content

    @raises(Exception)
    def test_special_case_where_only_one_model_used(self):
        model1 = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter = DjangoModelImportAdapter(model1)
        adapter.column_names = self.content["Sheet1"][0]
        importer.append(adapter)
        to_store = {
            adapter.get_name(): self.content["Sheet1"][1:],
            "Sheet2": self.content["Sheet2"][1:],
        }
        save_data(importer, to_store, file_type=DB_DJANGO)
        assert model1.objects.objs == self.result1
        model1._meta.model_name = "Sheet1"
        model1._meta.update(["X", "Y", "Z"])
        exporter = DjangoModelExporter()
        adapter = DjangoModelExportAdapter(model1)
        exporter.append(adapter)
        reader = DjangoBookReader()
        reader.open_content(exporter)
        data = reader.read_all()
        assert list(data["Sheet1"]) == self.content["Sheet1"]


@raises(TypeError)
def test_not_implemented_method():
    reader = DjangoBookReader()
    reader.open("afile")


@raises(TypeError)
def test_not_implemented_method_2():
    reader = DjangoBookReader()
    reader.open_stream("afile")


class TestFilter:
    def setUp(self):
        self.data = [["X", "Y", "Z"], [1, 2, 3], [4, 5, 6]]
        self.result = [{"Y": 2, "X": 1, "Z": 3}, {"Y": 5, "X": 4, "Z": 6}]
        self.model = FakeDjangoModel()
        importer = DjangoModelImporter()
        adapter = DjangoModelImportAdapter(self.model)
        adapter.column_names = self.data[0]
        importer.append(adapter)
        save_data(
            importer, {adapter.get_name(): self.data[1:]}, file_type=DB_DJANGO
        )
        assert self.model.objects.objs == self.result
        self.model._meta.update(["X", "Y", "Z"])

    def test_load_sheet_from_django_model_with_filter(self):
        reader = DjangoModelReader(self.model, start_row=0, row_limit=2)
        data = reader.to_array()
        expected = [["X", "Y", "Z"], [1, 2, 3]]
        eq_(list(data), expected)

    def test_load_sheet_from_django_model_with_filter_1(self):
        reader = DjangoModelReader(self.model, start_row=1, row_limit=3)
        data = reader.to_array()
        expected = [[1, 2, 3], [4, 5, 6]]
        eq_(list(data), expected)

    def test_load_sheet_from_django_model_with_filter_2(self):
        reader = DjangoModelReader(self.model, start_column=1)
        data = reader.to_array()
        expected = [["Y", "Z"], [2, 3], [5, 6]]
        eq_(list(data), expected)

    def test_load_sheet_from_django_model_with_filter_3(self):
        reader = DjangoModelReader(self.model, start_column=1, column_limit=1)
        data = reader.to_array()
        expected = [["Y"], [2], [5]]
        eq_(list(data), expected)


def test_django_model_import_adapter():
    adapter = DjangoModelImportAdapter(FakeDjangoModel)
    adapter.column_names = ["a"]
    adapter.row_initializer = "abc"
    eq_(adapter.row_initializer, "abc")



from nose.tools import eq_, raises
from pyexcel_io.service import date_value, time_value
from pyexcel_io.service import detect_int_value
from pyexcel_io.service import detect_float_value
from pyexcel_io.service import ODS_WRITE_FORMAT_COVERSION
from pyexcel_io.service import ods_float_value
from pyexcel_io.service import throw_exception
from pyexcel_io._compact import PY2
from pyexcel_io.exceptions import IntegerAccuracyLossError
from nose import SkipTest


def test_date_util_parse():
    value = "2015-08-17T19:20:00"
    d = date_value(value)
    assert d.strftime("%Y-%m-%dT%H:%M:%S") == "2015-08-17T19:20:00"
    value = "2015-08-17"
    d = date_value(value)
    assert d.strftime("%Y-%m-%d") == "2015-08-17"
    value = "2015-08-17T19:20:59.999999"
    d = date_value(value)
    assert d.strftime("%Y-%m-%dT%H:%M:%S") == "2015-08-17T19:20:59"
    value = "2015-08-17T19:20:59.99999"
    d = date_value(value)
    assert d.strftime("%Y-%m-%dT%H:%M:%S") == "2015-08-17T19:20:59"
    value = "2015-08-17T19:20:59.999999999999999"
    d = date_value(value)
    assert d.strftime("%Y-%m-%dT%H:%M:%S") == "2015-08-17T19:20:59"


def test_issue_8_1():
    # https://github.com/pyexcel/pyexcel-ods3/issues/8
    result = time_value("PT1111")
    eq_(result, None)


@raises(Exception)
def test_invalid_date():
    value = "2015-08-"
    date_value(value)


@raises(Exception)
def test_fake_date_time_10():
    date_value("1234567890")


@raises(Exception)
def test_fake_date_time_19():
    date_value("1234567890123456789")


@raises(Exception)
def test_fake_date_time_20():
    date_value("12345678901234567890")


def test_issue_1_error():
    result = time_value("PT1111")
    eq_(result, None)


def test_detect_int_value():
    result = detect_int_value("123")
    eq_(result, 123)


def test_detect_float_value():
    result = detect_float_value("123.1")
    eq_(result, 123.1)


def test_suppression_of_pep_0515_int():
    result = detect_int_value("123_123")
    eq_(result, None)


def test_suppression_of_pep_0515_float():
    result = detect_float_value("123_123.")
    eq_(result, None)
    result = detect_float_value("123_123.1")
    eq_(result, None)


def test_detect_float_value_on_nan():
    result = detect_float_value("NaN", ignore_nan_text=True)
    eq_(result, None)


def test_detect_float_value_on_custom_nan_text():
    result = detect_float_value("NaN", default_float_nan="nan")
    eq_(result, None)


def test_detect_float_value_on_custom_nan_text2():
    result = detect_float_value("nan", default_float_nan="nan")
    eq_(str(result), "nan")


def test_ods_write_format_conversion():
    if PY2:
        expected = ODS_WRITE_FORMAT_COVERSION[long]  # noqa: F821
        eq_('throw_exception', expected)
    else:
        raise SkipTest()


@raises(IntegerAccuracyLossError)
def test_big_int_value():
    ods_float_value(1000000000000000)


@raises(IntegerAccuracyLossError)
def test_throw_exception():
    throw_exception(1000000000000000)



"""
    pyexcel_io.plugins
    ~~~~~~~~~~~~~~~~~~~

    factory for getting readers and writers

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from lml.loader import scan_plugins_regex
from lml.plugin import PluginManager
from lml.plugin import PluginInfoChain, PluginInfo

import pyexcel_io.utils as ioutils
import pyexcel_io.manager as manager
import pyexcel_io.exceptions as exceptions
import pyexcel_io.constants as constants


ERROR_MESSAGE_FORMATTER = "one of these plugins for %s data in '%s': %s"
UPGRADE_MESSAGE = "Please upgrade the plugin '%s' according to \
plugin compactibility table."
READER_PLUGIN = "pyexcel-io reader"
WRITER_PLUGIN = "pyexcel-io writer"


class IOPluginInfo(PluginInfo):
    """Pyexcel-io plugin info description"""

    def tags(self):
        for file_type in self.file_types:
            yield file_type


class IOPluginInfoChain(PluginInfoChain):
    """provide custom functions to add a reader and a writer """

    def add_a_reader(
        self,
        relative_plugin_class_path=None,
        file_types=None,
        stream_type=None,
    ):
        """ add pyexcle-io reader plugin info """
        a_plugin_info = IOPluginInfo(
            READER_PLUGIN,
            self._get_abs_path(relative_plugin_class_path),
            file_types=file_types,
            stream_type=stream_type,
        )
        return self.add_a_plugin_instance(a_plugin_info)

    def add_a_writer(
        self,
        relative_plugin_class_path=None,
        file_types=None,
        stream_type=None,
    ):
        """ add pyexcle-io writer plugin info """
        a_plugin_info = IOPluginInfo(
            WRITER_PLUGIN,
            self._get_abs_path(relative_plugin_class_path),
            file_types=file_types,
            stream_type=stream_type,
        )
        return self.add_a_plugin_instance(a_plugin_info)


class IOManager(PluginManager):
    """Manage pyexcel-io plugins"""

    def __init__(self, plugin_type, known_list):
        PluginManager.__init__(self, plugin_type)
        self.known_plugins = known_list
        self.action = "read"
        if self.plugin_name == WRITER_PLUGIN:
            self.action = "write"

    def load_me_later(self, plugin_info):
        PluginManager.load_me_later(self, plugin_info)
        _do_additional_registration(plugin_info)

    def register_a_plugin(self, cls, plugin_info):
        """ for dynamically loaded plugin """
        PluginManager.register_a_plugin(self, cls, plugin_info)
        _do_additional_registration(plugin_info)

    def get_a_plugin(self, file_type=None, library=None, **keywords):
        __file_type = file_type.lower()
        plugin = self.load_me_now(__file_type, library=library)
        handler = plugin()
        handler.set_type(__file_type)
        return handler

    def raise_exception(self, file_type):
        plugins = self.known_plugins.get(file_type, None)
        if plugins:
            message = "Please install "
            if len(plugins) > 1:
                message += ERROR_MESSAGE_FORMATTER % (
                    self.action,
                    file_type,
                    ",".join(plugins),
                )
            else:
                message += plugins[0]
            raise exceptions.SupportingPluginAvailableButNotInstalled(message)

        else:
            raise exceptions.NoSupportingPluginFound(
                "No suitable library found for %s" % file_type
            )

    def get_all_formats(self):
        """ return all supported formats """
        all_formats = set(
            list(self.registry.keys()) + list(self.known_plugins.keys())
        )
        all_formats = all_formats.difference(
            set([constants.DB_SQL, constants.DB_DJANGO])
        )
        return all_formats


def _do_additional_registration(plugin_info):
    for file_type in plugin_info.tags():
        manager.register_stream_type(file_type, plugin_info.stream_type)
        manager.register_a_file_type(file_type, plugin_info.stream_type, None)


READERS = IOManager(READER_PLUGIN, ioutils.AVAILABLE_READERS)
WRITERS = IOManager(WRITER_PLUGIN, ioutils.AVAILABLE_WRITERS)


def load_plugins(plugin_name_patterns, path, black_list, white_list):
    """Try to discover all pyexcel-io plugins"""
    scan_plugins_regex(
        plugin_name_patterns=plugin_name_patterns,
        pyinstaller_path=path,
        black_list=black_list,
        white_list=white_list
    )



"""
    pyexcel_io.readers
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    file readers

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.plugins import IOPluginInfoChain


IOPluginInfoChain(__name__).add_a_reader(
    relative_plugin_class_path="csvr.CSVBookReader",
    file_types=["csv"],
    stream_type="text",
).add_a_reader(
    relative_plugin_class_path="tsv.TSVBookReader",
    file_types=["tsv"],
    stream_type="text",
).add_a_reader(
    relative_plugin_class_path="csvz.CSVZipBookReader",
    file_types=["csvz"],
    stream_type="binary",
).add_a_reader(
    relative_plugin_class_path="tsvz.TSVZipBookReader",
    file_types=["tsvz"],
    stream_type="binary",
)



"""
    pyexcel_io.database.common
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Common classes shared among database importers and exporters

    :copyright: (c) 2014-2017 by Onni Software Ltd.
    :license: New BSD License, see LICENSE for more details
"""
from pyexcel_io.book import BookReader


class DbExporter(BookReader):
    """ Transcode the book reader interface to db interface """

    def open(self, file_name, **keywords):
        self.export_tables(self, file_name, **keywords)

    def open_stream(self, file_stream, **keywords):
        self.export_tables(self, file_stream, **keywords)

    def open_content(self, file_content, **keywords):
        self.export_tables(file_content, **keywords)

    def export_tables(self, exporter, **keywords):
        """ read database tables """
        raise NotImplementedError("Please implement this method")


class DjangoModelExportAdapter(object):
    """ django export parameter holder """

    def __init__(self, model, export_columns=None):
        self.model = model
        self.export_columns = export_columns

    @property
    def name(self):
        """ get database table name """
        return self.get_name()

    def get_name(self):
        """ get database table name """
        return self.model._meta.model_name


class DjangoModelImportAdapter(DjangoModelExportAdapter):
    """ parameter holder for django data import """

    class InOutParameter(object):
        """ local class to manipulate variable io """

        def __init__(self):
            self.output = None
            self.input = None

    def __init__(self, model):
        DjangoModelExportAdapter.__init__(self, model)
        self.__column_names = self.InOutParameter()
        self.__column_name_mapping_dict = self.InOutParameter()
        self.__row_initializer = self.InOutParameter()
        self._process_parameters()

    @property
    def row_initializer(self):
        """ contructor for a database table entry """
        return self.__row_initializer.output

    @property
    def column_names(self):
        """ the desginated database column names """
        return self.__column_names.output

    @property
    def column_name_mapping_dict(self):
        """ if not the same, a mapping dictionary is looked up"""
        return self.__column_name_mapping_dict.output

    @row_initializer.setter
    def row_initializer(self, a_function):
        """ set the contructor """
        self.__row_initializer.input = a_function
        self._process_parameters()

    @column_names.setter
    def column_names(self, column_names):
        """ set the column names """
        self.__column_names.input = column_names
        self._process_parameters()

    @column_name_mapping_dict.setter
    def column_name_mapping_dict(self, mapping_dict):
        """ set the mapping dict """
        self.__column_name_mapping_dict.input = mapping_dict
        self._process_parameters()

    def _process_parameters(self):
        if self.__row_initializer.input is None:
            self.__row_initializer.output = None
        else:
            self.__row_initializer.output = self.__row_initializer.input
        if isinstance(self.__column_name_mapping_dict.input, list):
            self.__column_names.output = self.__column_name_mapping_dict.input
            self.__column_name_mapping_dict.output = None
        elif isinstance(self.__column_name_mapping_dict.input, dict):
            if self.__column_names.input:
                self.__column_names.output = [
                    self.__column_name_mapping_dict.input[name]
                    for name in self.__column_names.input
                ]
                self.__column_name_mapping_dict.output = None
        if self.__column_names.output is None:
            self.__column_names.output = self.__column_names.input


class DjangoModelExporter(object):
    """ public interface for django model export """

    def __init__(self):
        self.adapters = []

    def append(self, import_adapter):
        """ store model parameter for more than one model """
        self.adapters.append(import_adapter)


class DjangoModelImporter(object):
    """ public interface for django model import """

    def __init__(self):
        self.__adapters = {}

    def append(self, import_adapter):
        """ store model parameter for more than one model """
        self.__adapters[import_adapter.get_name()] = import_adapter

    def get(self, name):
        """ get a parameter out """
        return self.__adapters.get(name, None)


class SQLTableExportAdapter(DjangoModelExportAdapter):
    """ parameter holder for sql table data export """

    def __init__(self, model, export_columns=None):
        DjangoModelExportAdapter.__init__(self, model, export_columns)
        self.table = model

    def get_name(self):
        return getattr(self.table, "__tablename__", None)


class SQLTableImportAdapter(DjangoModelImportAdapter):
    """ parameter holder for sqlalchemy table import """

    def __init__(self, model):
        DjangoModelImportAdapter.__init__(self, model)
        self.table = model

    def get_name(self):
        return getattr(self.table, "__tablename__", None)


class SQLTableExporter(DjangoModelExporter):
    """ public interface for sql table export """

    def __init__(self, session):
        DjangoModelExporter.__init__(self)
        self.session = session


class SQLTableImporter(DjangoModelImporter):
    """ public interface to do data import via sqlalchemy """

    def __init__(self, session):
        DjangoModelImporter.__init__(self)
        self.session = session
